# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
astronauts
----------

This is a sample dataset build into the engine, this simplifies a few things:

- We can write test scripts using this data, knowing that it will always be available.
- We can write examples using this data, knowing the results will always match.

This data was obtained from:
https://www.kaggle.com/nasa/astronaut-yearbook

Licence @ 12-MAY-2022 when copied - CC0: Public Domain.

To access this dataset you can either run a query against dataset $astronats

`SELECT * FROM $astronauts`

or you can instantiate a AstronautData() class and use it like a pyarrow Table.

"""

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

__all__ = ("read", "schema")

_decoded: bytes = None


def read(*args):
    import base64
    import io

    import pyarrow.parquet as pq

    global _decoded

    """The table is saved parquet table, base85 encoded."""
    if _decoded is None:
        _decoded = base64.b85decode(
            b"P(e~L6$BNybrs|~Oclxn6$BCh04TLD{a{=iMRfpb(}W`|(2B<B_?qmLD2aUv^ud+^;v5a0B7SsesT*A3Tv-Zxw@=&V{sUCKhEV1XE&?h7CjyTvrQWCg24l>@Zl$7BgPFkJsli0x_D<HKw@GkG?A$MP@#pw|FHRdw2R7x8-}l<!FSj~)ZR-L)f>P7p24gJY_w`<08_a888_Wi7tEKSYN*e|$y;j;_J{Ty_?k+_F*xF`tfmfvOz!)pQpI2(ypi8a&%V3Nx*fwd??U)4^V+z_z(FRkYfQsBifu)u5RsyqG(3+LNgb-Y7iL3-#UR#5ga-$+EfyrP%%j?4!OYoq?n1qbK<M&p>7<1rulK659x7(*K&81ce?{mG408;C3)Ea>?mT*b>3CzWV2i&~^skhw=*^DPJ8`w)-_CZ>`C(zRRj9#gE0+W&W1!F7#w@}<)A&jvE%zvQtUa^|9)nJT0klJ#ka&HOD#{;FbB`_D5px4sW`=z*K6POORHUwh~0WK`yd+VIr_DVI3F#*;e10}|&0zy*`E(dJOtqR7N!h8C!2}}tCgo0m-CNLrFZEe9AOK9au6POWiOM8DI#+brIp&SAe!fU@$SzK5^sd+yFQ-abf6|sad_F!uN=c;eK42-b_uUSf|l!_jK`M_O=F~$HF78dYb5V)29a%<cEB5tl+!5DK8!L~|8{4mBEw70)YTf-P@;C`h>U?Q+mZjC35u?4NwYwvjkMYgrfzSIbQsUh&;!Ada30&u?;1f~L=R!%uE#uVCJr4j<OLARAY7-I_BUxdJ%VDG!b7)wx!MhMIYsmnTlFvb)D3D>raIu^4JiY#?tI^278e6NEsM!@@ZV2mZ;N1*iDzX`^e0MukD<<)`tXwY6&>cDMw9GDKRt!wJ29ynG9<^&l$?Emt!u5F1S##jTXZN;_q4a}!<=lC0#3)pyij=&gO;HHVcfl@XYU@BJHXEe9P=r=GI6n$p6H!vMCk~c6PRH@nNZ4FEZgT*{>6mDQPP<o~Mw}I(6c&)a9sd&%%A=sN~cx~WX0I(V@y*66wYXkGKU8!gTGy9D9+nU8_ED=8glVSH6m<{f`<lCUKoO1@|;<^0>S|a|Z5yqH<)@mQpI|CDe&6ot+)EmZ_0tV%vb|a|BF#}V9mPlbOQ8NQ`vCmm=8JG`jey@+SreDSYtxQn{=F^H<U@lN<+78_1vUqP>cOAG=Ynyt3Il-%)dcAEOcf0hOkrtQ?Tz@de8oXyIFe6dxzooD$g&1QDUMX#Hw<j253f5vNFrP}rQeZ28W?zAsz-9~sr;*iG5%qpX9mc2vTOXJVj4=eu|5nt>z!+1gbSZBMP$@MvcUukqzNYTpFQr)oP?6=|U_vez{sz<WM@`(Y;J4EKi+WzM1I8Q7hKqZH?b3S#(_xG~ko4Od1=rk~H<%ChZsWx^=U?h+Yx|bSo2TAjI?{Wk!mYuSEJn9Am=0c_V0mqA-%{r_rkXJdT$yU{@((~;MyUq1_0kklT_AYj##Dqcw%|co0byzhHa7t(6%P#p4VVgSy?*+oTnzha(Tp=-KB$fz##jQKmfsmL7tncoZ8FP%=}?&7x6zgXTjREqGGIzD`jfLByxt$w+M4xXKK%Y5Xr104ObO~M<u$QX^!Rm>7F5I!CPQj=AXJ)0Zy(I(4`{|HP-<-oF8{WVwyiY6-S)v;NV=ckSs9`SbD`}|^k7QfV+<>zr#0)*+?(jZY@FLBwKma%8(E)__!H6Nl?pct5k0655+dL&>OCTQuyxu86BicnYE$!j`d~r^aHgV2{~^4&V(0Y1e7r@W+*>{AgXuthsTJbL@@}Op2h#yT#+?RZ3;_b;4rb#)l}ia)BHgJicW^EIU@E^m%0F#GKzP`TJD3Z~61ih(<Z5j*+Z;>;|2_w5`*-Snu62|Bi2+;N1|ZE&0Mjp}_fZi${#SyVTZg^Z!AxlN+^(|@W)lEr0xzj|GVMb!C4X@UrZZXol!_s^trVm{Yy6-Duvxz`1k)j)Ir9ryw6+k;hox4GLblay`@_+cKQNoO56lU^-cRVn{rkXNpuSex?zq~AK8$&NU?vc=fC-^e?)4Wi6A4l)#aRx?wzr(N$#Utv(kuo9x71`gaJjXI#~FWjU_P2H7k6Mn&#mP^)!c#UD730n+<}R32WG>w<9g~T<hN?<z-&-rOr_LwO7k?a19O3`PFZ$fKBneRNDj;g*ZIQ|@jI+T_XGDjFdNujUDej1&Vm?Y0cxp~0fnh8UjTC=z3*2VV!?faIeAcNK4DCPO2r*dFd>)^)M_11P-^gmTi>h3lUE1zaa#<F3b*E2QR4~b<kta#*EDTpAA8%SCYX+;{8Dcxn2LwTJHdP`WhYbX1hawf?Mir9Ab2baUaIK?^MMbngge1GkAMli=Lnb+D$<RBSELTF16OK|fa&lm?nuS-o4{TPTQO*9fHN!k)vR~GbTm6)D$p&;;T0CI=~o~yCz#Odq?AMT&d`!p2h8XFPMm?<wPqkp35Hv<AWTKZEzw&K!kj#K&$oHR46m7e7mh&ai*i?MWgo<^)k=JTJ_s{1HU9;2&-_vj^|*o{ObCXbf-oOTJ+GIgAa$VB?{k@b;C@?M=$Ca$SP1jM=iJ&tk`_W;N`h}+r`N}M<#i!UM8+c!uGXvw!c<_h{JG}IQapk%CFnS6eTG^h-i+g{_qd`W$Qfe|^q;8+!h8h4{OBo!>0qRg0*~5~p!4g~OtB=+C76zLm#C%iJ2h7>!Hj^f7t%~|2_^)kcc5~_{|s1(C72MrT1&uymIM=l)MLqdI9|#AYvcB)iC2ON6>v$Ct^||umPh*$%*SPMt<58t&U;{`W{+SttjQyok_)%SR0PxECwUvebVRgqJw$CY8o_*;w}=@mj^E<?deGMh=3*Ob1QT&pznBJ0M$?RIz=SN;19yYt))p9U4VVhvwv}zbL^M#RqK4~r9RwOM9lYBEf{mzBf{w#L6m(3f84;Q!X$(RLBzP2vB2C+U5<uYF-l8f=Yos9{i`Q7iZsbB*P&C0l_A6i83tHOFH&=tey5#0DLO8VfJc{qZh~1N?b&Tl6!WXO#y@$y~L0$8A7`h(Z{n>Ud$xLVOXRDl2WCK|=$^2lkrusXq{Vzl%W6?FeCN!!^k6riRM`@r$A`s44nmbH!OkDw^-hJ{yI}sb(VMXK)?;yh%+MJa@>-v<bTEFo=K^d}6ltqHy7GRqZdXf2&F(GZHC+4xl(Lgl@Mes)iEEV2P3p=Om85jv8sUVkU8Ge(&byRcja{tiz!P;C1QE{VaJi_GZvDdMWXiVimMKs;nY#psQJ&_}JFbR^pnK(}rjTG+4Z};JA{2Fbz+pfeh%TQILpu%bCx9tu{WwI#)>=<I_vr^IhUcuWQsk`N-y7(MO<h(R6{6Tu}v#h2LTGceVUNu0j|5aLDW9VRz6{;I#>5YeUJ{q6^>$0TloH+Go0?(PY{!)m*gDm-K&H{j83L{2@TcL!=FoULC&Bv=ge=*B17mCbfNF}N#ulZK&c0lk+E8(@?!Xr;5^UB;rLHJ^aStym75euS!dJ#9RaFxR*9^ML}RYbMI9P_0VGKlyfOZM+g6Xkg6;LugF*vx|2#YT<i{>@w_xwZF2Opg>X00>TZ{Vt>Wgdr~!3bjoj(oM@2crd$=CThzGrNd5T)$-umw6hk$%J#W)L{RgYhwT0{i=V&T!fR8el`H4Q(#aG-w3K2rz#3H&o-(l@OmhvY&m5Hc3@XQ~2oXHmxUeL1X37~Dct?0g)c}s2GsKq+<QpVEp~ct%j!ae6V25fl#T*22a-o%Et>Xu%bjRYUn5DW09+i-NB(67wVe^YbMqSl}lvZ~7gWn?+iYr)vyu$!}mV-(GE%26yUa%hD%HT}NY+6ygy~s#amjO=Q6DQ4l^0AU8M0`cURG?Hsei5w)Kp#P<&L|Vl;NV=y^Rn*&rgFl2ICCm_P5*)B4vQbQbf)~sOz6@a`R;(NhT&k<a2QwQ8Yn%7l&f;vku+fHgvFu~7QpX16j@fbs$S5iY^gJoDT5d_;-G0%1}sV;@eDT9rR@P9i=Kq?f16Ol&Y7GU&JzI8L{1h*FNtH~?afFv)h<_T{3gEQ!b`p^(*x!9k)t6JP~}>fl#-3_e}gCy?L#0JDuHrVUgl^iS=^qkVLkSWYpJX`2#P(6`jHJv#@U;(vtX>b4iOLr`srhbJ%%P%)<3M*S1H^+OL!CTaYFHH7~`eEq|8Oz!YyLaro};g%?WueO;4<x>GGzvL3b3Kjs4^Zn=t|-Kva<#Y*HnIujjY2R;B(5Toose!2(GpL$8fC+J|;(GUi!~T?<XDz&u3=&oHMfmM7DiR3rh+pzS@=Ly=85Aij^BjZThg^})gm#q%5K3Vxjj<k&+ekJU{q-f2P`fPP!ppY&^@*s6K{t#H~vEbc06ducAhKf?|RJ(Ko{Ap!pc_|uPtzfm|E48(I?7=4F^Tc6f*SS<!ml2Gr|gYc?$5y~`6i<j4>?1rhtEpR?6Cly22fW~>9bvyCC#o!VjzQ+CJDy^R8nmLPh=3k6h8vCxQ=7P1|8b~{?&plTv3Ye=lwE?iJC5mn1uh!oHF*b}a6T(zmSXVW9RPD&GC>%@VN@0Y;2oPla>UxetgeLqT+)hbrmBKS9xL^wU_j`9n7c}GKb1JtsF+(#&B3qk)bHkn0J1SlJcxnYT-l3uTK2tMog7IAtOP3BJO>~KU#AN6b1vlcI+)Y$U7Zl-9w}XfNxY@HrK|UsP1&saTYxV+7lEU-JdrQQm8R<hh-^%ISO<@!*Sv5<*#esRNSIFW7RNc1k0E;TK<bQJ}aN&?72U;a`gfZ(?2+ORAaSA$()R@)YdE!tj$Y(1R02RCj70?DO70Lw_1QiAq1{^j3C=Xe7Z*Fd7AVe-8Lt$rZWf&AeY+-I7LM|XvXk~C=a%3P%axNeM001bpFa2Pi0Pzd}0{{R3$^`)lTL1zG7(k$afddE<C|JOtfrAGSB1o7(p@M}A7&2(sz@dYO4<JH_7(t?hi4!PNs93?Gg^L$3V#t_5qlS$eICALN!J~(dA3%Z#8A7Cpkt0ZwC|Sa!iIXQ#qDYxSrHYj+Sh8r@!ljFsFJQum8AGOwnKNk8s9D3Njhi=c;>ejpr;eRFc=G7k!>5m*KY#)W8bqj&p+krgDO$v+k)ubDB1xJ=sgk8jm@;YF#Ho{~PoP4H8bzv<sZ*#@sanOVm8(~<V#%6CtCp=>xN_;*#jBUEU%-M18%C^{v17=RDO<*@nX_lmqDh-Zt(vuK*s^Kc#;u#TZ{Wg-8%M63xpU~!sawacox6AN;>nvwub#bo`10x7$FHBie*gmt97wRB!Gj1BDqP60p~Hs|BTAe|v7*I`7&B_z$g!ixk03*e97(dI$&)Bks$9vkrOTHvW6GRKv!>0PICJXU$+M@=pFo2O9ZIyQ(W6L{DqYI7sne%Wqe`7hwW`&tShH%~%C)Q4uVBN99ZMDf001W3LL3zg88rj|1{oLxZeeX@6$}>21s0Tj7Suu}jyom@95w(b4_S6^Zf<2DL@pphVP|Y*7!*NlVQwHoE+A8AWpH6~WFSg%E+7CIEENP51Qh}R6#x|k6#@VN02KrkzycMR0ZbJr6$BCh04TLD{UD$n0s!dL2LQmstFOz>(7~$NyvVw-uFS@#tp@-A00000696C!$Oke5SwKFJ8OQ?ify_V_kPl=AvVeRbGmr)3TS?XeHx&RCpad1P1S}QG1r-Dp1{DSz7zo)100000000;Wrw0H4000007B(md*#`gs000007zn2a000000000004TLD{a_FPfd~LN00003{T=}W{u}`V{ty8J{vH7bzW3i20R!$h0SEry`y2rSd?Wz_{uluV{=eoX0R{g4Gyw(v?jivNRUn8VNExOaWD+Q91$5XLXO-s8N=LBQz!EUD2P1;ezyRe70gPc5DN(=`TV!!0*-Ud`lqe2*3zR{gdd(cD+7ib=GoU+dO~ht<5}i@jY_UleUNk|(=30vc0f^u#uo1;hc|jm%Oe_$>S-@c|Eb>%gj6eq0Gbxk?0fg_w_FxF2AyEvN&%j1niv^xVqA?9QMvQ~-85o5qroq)rVQ(n`<Q#3V*iwW|h9(zVE5UY3GY_)HA_@d4*Hkzn!UiIH!6X<*gXTaLqj|{y0000chD#h31Q|6100tQt1bJm)auo~~$^{mH2o~%HCelSFnnWBJ2-yb!0000002m0T2LJ#70000MHYf<$2LJ#70000O2&V@C00000000>*6$BLo6#@Vi02Krk0ssI26$BN)0u_Et6(|)15&!@wwJ-f3ptS)2<N^}_3IGfM0{{j94*(GW5&#PT0ss^M2>=cN2ml8F1pp8L4FCZE1ONa4000006I22K0|6BP6`%wav;-^_$^{h!6$TXs92f``000000000O2mt^90000002VeV2owMS0000002l}X0000000000001bpFa2N;0D%YqH~;_u1N|NW1O6NV1O5;J1O6TX2fp{;76Ak9H~|O#-}@W^1AHU_1O6BR2mZh2CIJQh{xksv{_Y|H1yvx3AxIgf9ApwGY6W!I7-yB{&PqqH*T51mv<D-C(7*uY3jvH_7AaA{6<cI+B-u=JVU#EidJB|6o_ft3sM-?8Kr^5_ZB4{xdlH>d)@-p!7G5+##O7Ly1ObTPDzFj7PI*BfW=t#)!dbv!EG+U=VT?cq*E1=U1_6ZcjP_tiq9IWXn9smQT8jmqMWQhcIYx|w@EI6|D5k;HOkr;+0puKQu-H<BO@<~HTPwkKN;40##v%#?Dc4juBEkkDd%+|aM}y`-6{C5{00000CYVzk6$BYI1ONsZ7zJl?Z*_1L3>L}-7XAko&;};tO(w`p92f``000000000O2mt^90000002VeV2owMS0000002l}X000000000087vh96$BLm02Kfg1Qh}R000#P6<!r^Oce+f1QGxMD77#BATN0V00sa606}APX?A4?0000|WprtBWn>5d002Z~V`X7;Wn>Bf002#4ZeeF-ZDnqB6#x~;0TtQ-EEUQH6$BLq6$TtO04N7iWprtBWn>rzL1T1jc4Ytn04TLD{UBr^0{{a6007Db0RkmdRV5`+Q$`g>7G)*9T;-IMWi_Q;RToufVFesKURQNlSCw8A0X0@tR}`An^ifc4MO~Gh6%$)lR#RP7MP6oE6=hX<TE=Brm7S$mMPYSTR7jFkg=_#+6#yobSR54$88rj|1{oLzb97;Jb#oOA7Rm({lmiyH119KICfrmUHUKCGQe|{$a%E%~20>$VX?A4*87vh96$BLm02Kfg1Qh}R000#P72p*Wjt)!}xCIpi5&!@wwJ-f(a0Kle018*&4K84`2QyI=CDUp)n=Ef%_T=2!(*{>4*E1D!7<EJ8qCQ#Uw!nTNlf`rh5zb2i&;ZZ?(*SQd*i6l$M%<+~7~8)ylf9V|Z!INbqVP}6ZxB8Mn^f0nPcF6NHz%WtvRk0vB+UuOXQ7UJGN%@EtzFi&+wBA#mSR&3-a51-8T-~g<1w%6Q)c^eTcph{tO~|&bZ;VbZ_i`rX}AaF9@b$nZl!3jGGh`Ot<pXj&(YaZkjef`=jBg02HSC4oWEA>b~{gRH8@k&V;x52;WkFArCP1TyfC}n0k=_AdJb!?tlj=>M&(v|{w2EY%UnJFiMXrIYf#n#Fxrzzy;G!Hy$Jg{ECy;cthH9G%EYXoOp4LItY4+sOT%Lkh7<CS#b<1$oADZ+*~*TCa$17ZP)zB^Y!T+Qc+`NovP|Z?$LQXZc|@k$+V@~CDWkcWjn$Lf>;-9G-J<%83C?45`+R4;IS$cj2X;j2EVs$p_N3^Zmob%i3vE+*F7q1h#$-kIBxkr4YgJg4*S<~rH0@?(AC}1`Y!<iM5zut2-E~|0qzpG<P^RX3vfBY5@z$ch+N`x@wgA6kau}D*X51;mYD|vX`V@-Q*lx2iT&VZ%tTo`Z0&@kqY{qVso}08Mu+gOaR(D!}J-rw$&SX$tC3fA3*^C?tYcfX%v>cd!N%nF0l#0P7%t>xiI!?PWFT`5EH@VpD4A6Pb&Tt?;3-YN&pQ<x&)L9q~(=i;0yT}{{<gzuR9op>z47pRLf2WSyvYVjk?hGboxkmR=Y)Q#=Tkh0ovMIY^nXSsbEo<Vlsxhm=a;iVO`PuCT3|Y&{rZ&x!vDbpl#GF>*-++AxmP2x@9mDB3Q{8iWo|F6M;8cW;i!iEJv&r4Y701<Bi|$Om<Br_>&1Px0I}l`6K9<{EceX0G*%Z7L>a!L9%*OO#Hx;MVdF#q?T@D5I+M4UGEGFeLG{+g*XJJ<b7VFIl%wlBTlJKVu*YSBR!<W=fQ+w^fycma#_{_|2G|q(en}X5ihBI*7*I&NnI{YQ)Hmt|EhBRlf-)3J!BJtSRsa&15=SopFmE@kD#k71X$6HAD!f~e+bJF@zitBirftlB2w^J}?Q-$V=aIfsOO3T$4jml($o@=$)l3m>x)s@ju%v<srRsa<M6}$!&&;~3O$^{h!6$TXs92f+M3;+Na1l_v-{}uo!1c(d(02l<_y8r(G001bpFa2Pi0Pzd}0{{R3$^`)lTL1zG7(k$afddE<C|JOtfrAGSB1o7(p@M}A7&2(sz@dYO4<JH_7(t?hi4!PNs93?Gg^L$3V#t_5qlS$eICALN!J~(dA3%Z#8A7Cpkt0ZwC|Sa!iIXQ#qDYxSrHYj+Sh8r@!ljFsFJQum8AGOwnKNk8s9D3Njhi=c;>ejpr;eRFc=G7k!>2$WKYsuP5;TZVAw!1{B~rAAQ6oo>AVrciiBcs?moR10w24zEPoF@A5;clcDO0CVrBbzuRV!DoV8xO(i&m{#wr=6drE3?jUcP<-3npwBv0}!KAxoxg8M9{2o<WNyZ5p*|)~;d8rfnOyZr;9u3ny+IxpL;tp-ZQ39VK?{+`WSrPu@Iw_3Yilmrvh5e*OIY0~k=?s)3>i7BqMeVM2uq88&qI5Mo4$6Dd}-coAbpjT<?3^!O2ENRcB+mNa=1WlEDO&#h$H(&bB-F=fuAS<~iCO*nJv+{v@2&!0fO1Qj}zXi=j_ktS8Tlxb6^PoYMYI+bcwt5>mR)w&e`001V8ZyXf@88rj|1{oL%Vrg=8XkTPubY&F`7Rm({o*fq86(-zgCahT;7zBt6000;S-Mat(762#&hztM#7zEwA|Nj6PEENP51Qh}R6#x|k6#@VN02Krk#s?Mt1WXlz0Tl!i001bpFa2QS0IdlCwh=cZ5UV<4jzJCEQEJR1<s4hNUp?Qv%YPRDAOfHQ0HAHE#c!MfEC4G2E&%c=98a`f15(gI1ud9R{f>zoDK?Ngc@{{WIDZ0)GA@3HAUWjks_1tQ{w~w|-6K`7-Cm?ujsIaOw%Mjp_Ccp*H*ErcfxWga{1MVty44BE-k*>W6LlNtRzIR*NVe|cM>G-isHkc@h_M)4*rW@Q2c7<cNM3a@NNxTA>Y1d5`JF=ARTs-<WSjCIsMM8YQ?YE?w0;66PQRmMbMXT>vSB$9NSA^758A}m`VRpsAmAdJn+IL$i)N84)|9xqImY4#7I<I;R@DWi&61Jtz={iQCJB!q@?p1a<wi!s)z-4s8f#c<jkT>EZ(_W_y?Zw2LPv?@j?!49)S`tBq6rsP7!?2&hy@j#1uPZH1r-Dp1{DSzHUKCLS7mc_AV+0#ZDDSC7y>~|0000ewJ-f(L;xWQ00RI30Lldd2U~!EU_dAs4v2(6u{baq4nyGqiAV|z$H1W|2qqN6qVfP(E|^EfkP$#61&v?>5HK_x4u-NQ926bT2V?<+1OR}8A`lfuq>{q{p@bTyiQu$&L?);T!;liB90ft4z!5l928Lr*C<s^rO{FzjF$4k!W59^G1h10-WjL^4E&_((D8vA%2Lr<Zj53E%OeSy)3J!sz&``u6iioN)2oVSejKd+ASQL|4fmQ=376hB)w%{lL48{U?cnBni++(3z000LC<sbl5I2H$B=U5yjApq~B0W1I&KrX@{kQf4$-~dBaMgYR1*YFV-g$bYl;BX8Q!J<N75h9MsM*(p#REixWWO<Y%gOTPW01yP6!eR1Z{9+Hu0tN(B4vm-rV8A#SI7UGraD*-~gM|TEG%ALJ0zd!&04B<N92E>1H3R?#87K>4X>xREUvO+;V`T+%bYXO56$}>21s22&7LW@j+I1$na2z%OC=6F+b95j_WpZs{Zh062K}-M{EENP51Qh}R6#x|k6#@VN02Krk!Za1!9849E1Qi4l001bpFa2O43?(@L%5-oOG;nFt0sQ~}|G(4Uek3DJCP_q6+MB1{Ev<~lSRfjK8r^2tMsJ-z`qvFJs)29+|6SBg0Vn|`0VV-V9r7%rg^zbeT77wFrs$|yY;boC>yPLYgAjBl?hXQ<asP35j*Mmt_1+Rrzvm;y-7!4(1aVWCw<uXHcgGNY>xsR$<?b3j8N;aKNO5-nP%%^6)(N$8cM6iX1I0-Mcen7YE4e#`R_^DAV!UV_QMpM6&&`Z5?#{ve9h87rF^{`jkQO=K2=30|xoN!^8$_BZEBT)tl-wNz(`^kh>ca;Z?xMuqH9G6L$tZGn3;4+`p{7A{v?vJ^Y-$E#b<M@yIlQ)iO5B|zg1lkaKd$BO95BKu!QCl9+!hMrp4=S+Z!cEL-8phsH|+7wFb;a;?gXGln#~7SBg@@AxZ<wouG}31)7BIf@fUZ;z(MASad!>wXUa*+-7P%boLq<x+#N$5S<GY_W2{Jw4wNVE4!{BiDk55XVcJ09?}K}CMTw`x?Eo=s9s9yOH#pfviOOiL&CsG8597YK<rTCwsOK|WjFiqw-44n0m8_d$F}j4YN=qeLPtno~C)6fz7Z*IrI?V0jzZ(%l))smvT9^=py?R>3xVwh*p7a+XcgNr?C#-0gkh@z*h|v?Q7A<TO74Z+5%)#hl>Wk~N6S<C2M4z?C-8K3-C-l|OtdApmLIQBVX2R%Ovz#7x2Y}@!#1F+BGNp!`)kSu+xVuKOx5jKH>k7rVn7+Us)!;=vj7|aJ(kQ~|=T>HLcMMZ5Ufi7mXFA0~&kd*Nv#qzGMMeAt<K`we!kJCM-8E1jT+Y(<p;&V^$lX0WYg^Q_OlF7e4>$qn=LWt;3xcu28jp7-JtxUndju(}AN1Y<kT*=UxYNH6)r%H{s9fF><@!+hjSF;MA#WH~Z%Ap|wz($_<N1j3C}Yf(;O-bOcT+Kf7^8+#B^2YursGk6Fq6^8@mV@f3%ki-oKtQ@JvRtVjpBmE&Et4l`{KvlDR4t6u~umh+O@&mIhsieEAH-rGk4YES%X@GJdK6=+!m;t@gJi|dtAM%j??3N>BAGdiJ=xf8G@S;lMZi;<3u?srYo!)k*4;knJ}WF#px34D0CPx`*>v5#flvKhj_vWVx3$etC@1a-8CX)#tpLKTG|_$(W{X5z_UGQn0;(U>$M`Sro6X?lPV>o>^~+>j+VPqXqb<qD}_o;fP@}(%!Cm#dTKhI*NEyQ^Q>494rx2%jbK-sX!X>z4DKdl?Q(Yy-g&5que6n@YX-!q6|D8-Nk}*6$xA|1^sO&8{dh;qTNF`j7gr^wZM`wM|BrSGMTzIs&~sP3Wh6$I;O6sSRA0bO?xT+umKB8*?~G^k=<s-F-Z4WtJF<9Z8YaY1lgVTS^yo!N*l|w@^$aKOu7N?Vy)oVxmgyOP=p;XlsqKWyjkMP!0=+2wh6(rN4a{kW7`;inHv+YZp*m^tjJ~*CQum&EqFB*@=SDB;IWt;-WZX@`-9bzk{UQeQe6(;r$PlNL5_oP_k((8RaoGVSt-iQMO!9WX>E4i5UsUeDN(&jMv~YsnlqRKwi)nFp4No_xv)dCDy#UwVY(iATr@EqB6q68Pf~Q0vfd;6N7=)N(Oc9BQBuSFiQG|qH6vZMR5ipR?Ds2Y+7&8N71!Mx&PdSH&q(WVz3}OMG1Z)@MR!6TioROMJ9L=XSB-mi{>Us%WB~@Jeg9@w^oDul3`ty%{X&AG-rN)oN_jVsiWf7Wo1%}=8maP(+1<8-dI74y3wJudz2)hDT+y^hgWm??VS-)%ltyMlfeE$bK9A!P!Iha*;O^c=FexO&t9RaWJ8bg%Ie9#m37#J)+Su|)Zk<BGmb@==@s>2$aKI|VVC3fJf;ih4#b5Cc`Y*4U|YCF;;*wmDgl|nQGrrju0P_rX8{O<wTVNPyjIyrt2f&4TK+8QBp(X+GOUDQT6Y}a~XYvBR+kXo%G)FQSNn6bZMk#5OQ3W@}=977>5zM(eYaA=v9@0B?^R{VJ*tgfV-k!q(|<|WSb<3J7E@-D0mmBS2`pN=<KjfR#!NqNNh469h9+Bxk`o5R}&{2&T-SUL5+iJqqjdW`qcvt*N=i)!rATeGy`6^e$sZ?;&VEt=-To8oF1XT-n(-S-XaTB?a5N!mh&CH}vykX0yoQui@{=9{5+V-E8IxuLB848_-iuQfE2m)goZL__MIL0||>h-)Q`;K!XE#^i;Nqm^Q;Jp&B83&ke%bT`1xx38O-(Y%kHFL|hN4T~$;nDLasr<k&wGA<9DT4-t?m!qdr56&g6F)=1F2+UrZ!<cvv!D0A2!xN_Qf0Ckus9dA2qER4bMb16<eIY?lff%?76#x~y1{Kf-EEUQH6$BLq6$TtO04N7pZ*FU4a&s66L1JlcXLN6F0000ewJ-f(o&fO-00RI30Lldd30nXH2pB-1fPn)D5-3=}pn-!25F!+iFo8k^3l}hC(6FHZhYlV-fCwRC1c?$RPM}Dkg2RLrELymD0V9Tt88m9xxPc>wjvYLD`1k=Nh>#&fiWoV9BuN1!OPDlq^5h7VC{m_Ssbb{{mMmJfaEZ``ikB}1z=RPqh64mLWw@9*gGNjmHEY<kaq|XF96597)Uk61PaY0>s_ZGlM@Jt&e*guNAxO+1LWQmvI)o^Zq9-~QF=_+`f};YDAVm@rWr3tgl<-uF!oX6KCJioOGQuH~rcE6<b@KEH2B=V?Mnxi3s>TjeGGs6~g({V*RXSL;vU0?VDN{N&xO&9_RxDYwXw|ZH%L=XzxlHM*va5-eAiR3{!t)DQFku{q$stzE7_Vc<k||rpteLZC(4qyAR%9BrYSyk<!3Hdwwrv_Za0As1f(&opz=abxj$Ao&1J2E;atK{IDmB!tW7kfGyLa&7$(u*7Ry`e2?A^nckBdGT`(*Iz=TE~wfQ$kP9N5iD6oL_S6f}4c0000co}mF86$}|Q1ONsZC<|g~a&%~4aBN{?Wdw9@cWxC77Rm({_B$4&BqsQo0Veu<95w(b2U%}!Yh`kC7zjaPX>Mn9Z*Bk?EENP51Qh}R6#x|k6#@VN02KrkBo#JH6$BLo5&!@wwJ-f35|ID^1ONa4O<`<h1^@s6MrCbbY-JSy6@md3lmRRi$^{h!6$TXs95w(b1WjRVWf%rVWo=<>WdHyGD77#BAVCQN00RI30Lldc0U`hc1u_5w1rh)a83SMl1Oy-e0|6WW3xR@Y2oWRz2SWjr5C8)cC;$f;0w5Lu1r;C=011&03IHaAr2!lj3>h^900tQt24`h%WMy&{3>L}-7LEcIvH~WgqX8z;q5&K>04M}aVQgg>21aFVVQggp87vh96$BLm02Kfg1Qh}R000#P6_Q>Rq9IHbqyrTM5&!@wwJ-f(i55LV0IF$Z5;{PuBpt&r48t%CW5zL@nbOFTUC>HN*0dy~NkT$ubqo#7%&stPY;E`f=ae8SYLZr;`@ZjEO3g|E69E+g7y-!PSeB>LY=N012;(sOpTKo;IQf7%te6`4Pic9Q?eHW=Byws_2mc3TI+wdSV!;$jAB@9HT5MM%R57)+d5*i|f5vuF&*IF1w4}B|7B@}dAs{uDbObeN_T~S?@|#vEhnfo${`gOkQ?|wz=BCx>(BuEZHF4vmq8woyX8EkA&+P}xsJVl+R{kG&Xv|cUVfoOKu&3uZgf%QdkO5A4wY86i7XCY!SQwI+{}3o?bL*hM5?*XgEBv1r`GEcdz}8sg(YDw*g$69Yj8;43|HM_Lgi(cEjZd3GEJNYH1WVZ5_)@|0;ma^we9O@P3tYyaU7^6nO{0b3Tg*JD+sx617sg?BJ$0QWx)lCz1bAUHu3YKLFb=b4Mf#%L<qO6F*Eu~)<bTA~{NS`P2|oDP<*q^q7AK`5h5ln)t;5`wNSBK8U5!KjA8=ZNn5V$<>uk;Hcoa13w|ODJk<)0gJJ@38e}QM=a%Y8QlOz>+uHIA}{}V`!mIL-2@`YuXvl5Yt{C9XRH5pPu40UyCt$-Z=4P1>7+h7%HYpR*DVPtDsa>5JYX({2;3(EkjZaH$QaoKX(3UeGGLC3=UcW6yLSW6P5F6<%0)y9n%#y~ff)g%mL2P3e?Y?o*GIg34*E|#>*?Pkb-0=il;hDlAre~XF#8Vw_Lc#<akD5y!PjPM^qD^!U`P1>6O7M;(f3H6#m{#&%%=F9&GZUO9S?7@<Hl|F5Rp4CQ(y+tbOgjlAT0nEh)ffZ{qTI0xnjAw(}g_i#iP9>&ZLIj@cq^in)2L%nK<xS(|*<O&17!ZVIM9R?!Ls(MJ0o#-sJNRGm{}+#Ajr@Ob;0Yt8E~_09_F!%)e(?X|Emxft7>C(&`2V7^rqzzKp#Knzidf_|-E>Ne6)L8%btF*uC`1h4NKX=CCRqV(RG7n)iHeJRq(ThKehiY6q`j#SC0x8b8ZvCyOl!h_3MKbq$ibmogg~i~npUe;3wy}iTg&6Y3OCp)*2b#*hoDeO!hed#05-lj!hZps&uvokKO&tKVPEdObQTq%a|lr=e9KNHgr02K>9jnpE6=prnXA)suuZGaS(OM4iiSO83%DCE%wiM5e+r5pu(yhtMiXp_so=B%3Q9U%im=tNba;^zTS^tFne!SwB7KYPRVeB#rrHDst1k?nl8~r#>YPFvQa(^qTbosh<LTx9#iC04@n6GXYlwL;9tm35e~N{h)?y3)DH&mFs9~!nQbUWw5Y{-J<taPk{{uemAckd7&*JJeUY*E#n-JbhZG`?m{FiuIOXx|Rayw(OdP+&l6&0ySHkYkIjBHdGugmUu)PybnCGdi$R190pXmnyr=Qc(Dvuy5Q*{N7|l9nLm|3gM#8x%>hAa<n$hoe$sE9FQk^k60ll6zGwKbIv99XwRD*p>$P*j?G$^8W!f?5euLG8WEpUjl)QsDUvth6s_&h$Klt(im}s#3<(=rg{+|Fr<`1N|_O|QjmhAFpL2K_h~lBl#`w5G_Qh0$eLe1sDocqvq+NMUUDy$Ffi7;ZuGeZdQZ~^e`K}4G;-=;iZ1jVoPM9e1G9yyEMuY4Hi=WWpICEws#D?=r15S*twqF^GiqXE@Sw5dK_<4dA?BEQ5umN8V!M>oA(=d6YD>b_;*68587Bn=Q5Yqo^$BJ%^E&SBR0SPovltycm+#7m@8X`YDIsNFE9S>EM@Hl=%oQ;HKFBug?ofCFkp_hD^556CJSQo*&Jc6j(J1n_s<0rtpfV)=U+g4UX|VNK0r#Cc;qQ7Y$tW%SSp`lWOc<<m5H{@>j)<y)H>0^=bDTP#!H-9?zsF0p(^FKP%rQWLaN+TUdajJ_qb%VHOrA24<+!s9p6kx%CVl&hat1hLyi4jE!H5#h=W#_sI+p{J-%SNQyOIk)M?j}vkceVlw&J}+;2WJXK7Wl%7=s`DADj#s82-+(V-sj&j{?2`xc=<unmMdREuv8@bQp2i66sz{CCQ)gHVCBR=s=evr(;wymAPj)dUj&9gnubknwMYyo~Gu}^0;sck6c4juS6?Zc>iW2WSQ-+4#pRFg$4PKv@!32$42vnt8hOah=E%i3#=o+*bH_~Co4ie_GD_XVm=B^2p3o=*Cd0Hizu<J&%m=Y3+2V=#-nr8HbScqDU6O9$#<fCc@Pu(b}jC;&>ud5q0u5FS4B*?8?d4ns>1Fk?fbi;Sq#f1mM~-!e0&{(wA6bNnyICp_BC6}Wo+QI@5r~wh1`_HTNXQmLLnRkZ<;8@NHzvhMSWm!s4GFQ2j)HsqP?hkd;otip+WO)-E+GVqvDUox;o**dpAc~ubHYsd27}r1T}N=sBtT1QOX4e69q_N7Gk(;DExXAVH!w-5)kDEOq4h(oS~ZrL^RqeL&e~fq7jzdHkN~OLk(3{87*m`1#S%`v}tzE#A!tjH}*{D8sSQM2%T9<*xaAIFS{?|6Q)(o-GbIx;Lw|wohtz^Z-i)^u^*HZkQy)S5h}{T4>UTIeN)Zf%5*&X{2t66Ts!e+7~E*Z9qTRDj58c&X2qqMy4-dStQYj9PCk(XJ!KOiHq<ggabd#LE~MRFNQTZv+M%*=Ud;w>a{-)P`-qg7UClVv0c^VI-q*p9nkdXrXLJBcksvr%N2Wk{MTyd+YdfI^6#x~i3>C-=EEUoU6$BLq6$TtO04N(-Z*^{Gb98TaZXi>1VRU66Rc>i^WpZ<Aba@y60000ewJ-f(76R240A{hZF(9f@2Ry)7vT*$eO&?{*wk{9db<;XU+qt3T)L3#C+*mUAc(L4EN%kW_tZ-5u*#WQsqX4J?vkjj8r(ye4Cco(n7+(L<1q&4}g$X^!zM81Bk3R3FB;ob1NofQK4;;|c21sX`stgt!r8}L1Tmn=hz_=41I1LoIdA1Q4P!oX705tOgn3Y%p?f^z*S>741B|de1a&<tJ)Wy{(DLKOFYYIu$b6P@%jj8$5>K=zz@~JEaSR0nc1(JnWvVE~uiwp^u<<!U!*$XLQUyEJ=nFL`=0r^i}tq@HdBv{)%R#}$BLypG%LZEOE9kB^Nup06-``L?9j|B2Ru->^;?|1K`$Lsuzs^YQ3b*xl5aF`o;L%*%|vrz-MW<NUbHf1!k-)d4(G-|`$;^%EVLkHD-bQWb~s9X#vT=XSB|H2ra*wO#}3l~Iohzhf}4&o`@Ic}l<Ml)IL6)G@Ons?5cDS{|^Ujpw5Cu{<jz$*!zb(0cy{Uc}@y&^(j5W5c^I&cU>HoCFB1_(in%TqPkGwMrob}<qlW9tpWTCf#G6U~_|8|Sap%h-j@aA{amT!zdab<R0&dmbZ(HN&(--c||R%K~Hk!UjNYb*o<d?2?byah^T3;<9Nt;g+Np>c!8}61<<~2F8L+ad!OJrTz+QvL<T<smc<SJVnZ%RZ(W2_WsQ718#Lgd2){q!^%k$ChU@J8!AbH^C598&+65YYb4*Y2$2tES(as4UO9U-cZ;why^TJ$Qk*W)m6v5x6A}3i9ZlAZ%1S_Os1PBa>md;9OZ@%sK|!FPK&%FYx|WgX>K{L;9UJbmb*K2;J$guR_?^guJ^Reu*|q4$?iTRexx#O!=z!~+$ab(f!?@LP7(;2Qv%~0{gWCh0Ih&~}ca-~DbI3undl`|}nT<o%GWl$*XzGWfM1|k1JtQ?Zd+Ku=zrFw_mec_p6$}|Q1ONsZI0|8GZDC(+VRU731Z-(@bOdR1Wo;D<7Sak9&TAI>EhgZ{0VbTL0UR~}C>vRCb#7;KbZ>WVAX9W<bY&n_ZfSO9a&u{Pc^Ciy87vh96$BLm02Kfg1Qh}R000#P74{z$f(%R*rU4ZM5&!@wwJ-f(`v$!i0EQ%JCIFM01HceMid}IFi;}pmbFP8|u<-=>do5G>)GpP5cY*mP0Zsv009^oJ0ISoL2X$HW-y$`JEjrc9e*uMNs3<r2A9-t;M$P{T>`Oi{uIAoe{!={pzmjYzw>9Zz{##lZE@~&ZGRawHozCU|WT2J1nmkf!{B|2lE&4yHB*)xp{Eq~VAwL*3$ri=!@?X*f{}+_0^%nDg(jiOp-youuVW$JC9_$BbqLnGvoJRR%g5($irP2_X{{hv5&D2gxz!?7{0|~*l)Jm(ZOm=VMju`(f5W2G3mUWX3<bPx}oW}9r($vDPHvcCE?HJ=Nj2+5h{!6N<<JvKvH2GxXC-EfZ0xTTDnEwZDt#<rBSZ%9uh6L)1N^;CiTD8Z2fhQZPx8SzLJn~89w^OGb1ut0vVJ7%ilpJPm^PkcJt8G~~l6)LC{|nkcBjt{n{{a6N-eOUgxy=6qiK&@MElNnz18s44{75LDtSAY!GHiwmZt}^FV8)sMlzfsch_<M>@xMU1xr7%PTX3N{hD#$S`X7)G1wYs^fRU)7IU_YAk|Ze<hA|*QVvu8Z0}-T{Q;s=RrzuEbID<Q}Q{vNUD{P_MA4HeJ$<oxRaBz_F=|ViDgA@j6&8|z31$7vcq<ELBW>drsWY79l&~SQ*Y~^WTG~y~zu|c^7>jA_nKL`RhQ~;LLypy5psWh|I0r5|J8Rf4KcT2RHT4s*!IDC$_W655Pm-qEB)`+J#n$NNgu*r@NY9L|vL^*5(XV$vYNMG%$OQCLigVYnjq&XjOSzC|+EK)Q><uBZZs*||n>Ol=UB~aX(4dl{n;7T~30+c3k>1;ER9u1EL_K1g_eOkUj#BlS!ECYb<Ws!_Ft=Aw+17mmiMCVjM0HEF+<5vDufzt6$xH-{g>WgbDRm!Pofd`qhU{EXnAjo^Yk(fBTge=XSHeq_l>g}L>Cj7oAYW42{X@;>NOdHN@HhjE9tLa&)6#x~$1r^i<EEUQH6$BLq6$TtOEGP$BZ*OdGXL%S3L1SZYb#8QNZf5`h04TLD{a|1K0Sf>$0000y0Ry%~0R`OqECB<)Bmo2T76AkPAOQpZDgg!l-yQ)2{u%)T{wM(h-Vy-={x|^x{ty8NRe*qCKp+$h2Sh@FFf15^1_9uBG!TsiBqDJvCJhIrP>D1whDC$WKvXUSj6;!G5HOL7qp@fZG>c8-!69Hei$w(EKsXQ%j^_iaghYsug9KP08U#YgbQ)n04hM?hAf!}*6{{gY2o@nh!GUQ-A_T-@A#g+t76Jm1SX5k%L7`E3IbaQkAm$>39E}aa;aFfsvc_!Wpp9CtMI+)sSRyss22;5S3JV0Tfk7a0x`hL^TZDL{5uztZU;qf;4;U0W5D1!sAV44*1cw7bvmg?J;b=qLISdPjqiAdpHb)E6wIC#wn@tFya4ZK4rlu(gITnjhLl7cV76fO2Kp;d22mu0tAS^a6#H-?n5DcLw0<l>*P_YVufPyw6QA7hFT~q)jzUToQ6$}|Q1ONsZ7!!4FWMy(^a$#h3VRU6*ZDDF}auo~~$^{m(CKl2XCXnI*ChXM#95yT{2U>4$Y;R|I7z#mSV{dhCbZKs902wS51Qi4o0ss{N6$BLm0000L1QqB@75o-V6`leW1QGxMD77#BVDu5CEdZ9Zm?D7iGzTGjS$(o?V0|d}x}0M#QobU{3}GX`1v-yYUe5(7i2#lOkpLzWs0z7l);;Khm9~jGKKHC9{THBDSDKmj0m1xNXczlA(?YJXuPEmOilhGykU6$mDlxgC!*U=){11roRi2f;_$~`(Ded?u7V}?3-K!W!IV*+vZ$RCvd^M^vX{~-Rm4f*nI>k4;<2%LCw&N6)vtj-_Q1@ttD(z$_P5%$FlZ3Ks{s#mf1rN&7HjQl7zGO}5Y;*F(Y|Cb)7_(B$iZ|P`+COai@9~mh+J<tIoA%DqPU-v)srJgCEd8g5DAzn&?SGeTZnTh}kDTt2{)7B)i1JlzRvK@%eegvYmL5<-PCT+%>AM;wvyd4=4I8<g&i{q~4BJF%?5j^4nY0fPkx}lU&8}GLk}}(7bqz=V1E~?qHq225lFynfvyFx(Oxd-7_Oq~RTK7&}c~(<sAe&q}GA#XH=!sQ_&1%|;Iz349UnRR_tO#P9Z4ngmn_Zdo-{QXlMs?vZkX=$lVIy=+`4jptDros){v%*?aBND2L1FJK<T$&=oBuc2l$EFSVVwV>5N#KIZQAxgP*Srkl#+dF6vFDuy^Ckd{0}wu)hCWjn&7B5t8ZeiO$#|^Tjgq6>xCXFtcO7wi1D;CoB!1Ug-3Rh@S@J5f;#tRqf%5}b+(Hv^B;lIf(GKMC5Z`TC{#NW1ke+!Ji9)eZ7kb5z7q<%)uQU;(SHf|q7;cv6N8Xz#D9+whgqZl0~spC^CUYmSMC1~kre}zgl#R^<oIUS?)mR=we<pla;U*Qk`WORkt9h9!x|w32}FUU*VF?Nh|)j_vLMN*Vnk+yBuUb4x(HMFG<{`c;`mTTCHST-nxwudDv*d{?~O`bC|i4xYG?`N9#R&qOw%5*@O=k0-)2z48H?_HBo<TVDQ%Zt5p6Yiz0~Pfn87@9@sIR6$ZOUTj`X}crMy#FZFIX?)ZS&~Yc3CRn^1E{g@+f2q2=^G4%7*s>>A%NQIqd=$$&BezcW+zYd#x?>6c5zJ0Ig#LxbI61oHj}&rT+ZH<OO1+a;e3&xcg%I8T6swwN+xbRPY2zF$OUDWz1CizzkE?>8uEa6N~K35f?WER;AVU3^YtJ}6F}`teY)H5X7`hiUQX{#mn6-tBg*j!=dW3*UH@_r{gZ;d9pSr+knATYzs=qD_BuXHJ+1+)ERcn`F^vTvmy~6nekmQ~Dq*r0xRC3U^K>_CDm@6&WjE$Wm?I&Cj}>)Unva)!-*|ssHVZnRjISi0f~Z19DHG)kcb_w_H>_t@h$J;ub-bffVS-vRn(P$Yhqfim_Y6D|jsmhkbKd?uwG&BE(3;{~GaT89B{QBc{jo<T<~c4AnT3&kJilg8T`>gabSL3ju-or^-|!H&n0kCU3^M6Ht4ieu+{lTP2`R`v3;(NMn}x?n(H6!OFtw#SQrZThyuFMWV+=M=yPO=S>2nn(lzOVVf@{eX3C`qiWoUj!?1?G-PvaJW8jw5fxV!_$bqHweFU2C2BrI_}va%AA}B2-*BL(@r2BnERIA-iOs5)<n8L@9MnBv8sG~_KkDK-igYl=-tjQzueO&-aK?w$UX;O!r!CXa<H@4~Z=~%4l>y)Qf^inMk=6iahp%Za3D`zQYPPQ~l_(8dr#RytLp%>s0<J6FqT@&~5bK@^<jnymp&$S#(64FbS8VocVG}cWlLi`)%*X-*lO%DX%y7%tEkh-#_1JAxxY-V#sRFSnjCn64YIO!IzFB!lERC*mVGqiMB0w{T`VqT4E^H>{IrTC_aW9!h*(~q*U~=3mh4Vp0pm$CV1G4q*;9Qh_o3V|8ijQ#MH$-3#9ti&z1sF7PYzjFB;n(T6vY%RQHV0n`6898LQ~?zL72X9E@&zmv$^{h!6$TXs95!|+DpqB5WpZh5VRCsOO=V<hV`*+>J0MVXVr*$+AVFkpX>MtAbaG*IX>V>AC_!a%Z*F0AbZKK@Y#=5eL34C+Z*F0AbZKK@Y#>E$XK8L_WpZh5X8-^ID77#BU~d3<3jjF)001Zf1I`oy4ff~z{eN5>0R#RT0S9WVw-f;f{+_)$0SW%@e)|>y1O5;J75?{azdd{R?)i=o0RsXEO#lG`0|W&I2M7WN2?h!R3k(en4i69p5fT#=0u>eq7ZD8w3<M1l7!3y*8WS5F9Rddp1Q#9$A0QzDA`uNE0wg6SCkX~91_miA5i2YfEiNw!3<NL&2QddSGc*Y`Ha9p2IXVXi5e+*86+Ar#0t`MsKm`UtLI)Z{4MZ9UMF~bnNCyH*6-rA?0s;{S6%9=XPESw*2Mi4YQBp@25eF4hR8>|7S6Eq40tZK02S!_5U0z>cVPXPfWMyV&252Q|Y6c%`2W)3;Mh0#N5hX?y3~vH0a1BOAadLAAB6M{J3U+sRXB8181`THi5qWwddjJUu2Q3Xfe0_dr5(gCmQ34ZxfPoQ$5rYN|77>JnTrGx&h>3~@8wUahU5gWp0*zb^jtUb10000004C~)0vr_#88rj|1{oL*XL4a=bzyX6Uu|J(Z*mn37Rm({#8VdR9wwZD0w(_H0US1VC@NNEbY*gBZeenHAWdatX=7<_Wji2Hbz*F3V<16fZE0?4b98cHbZKvH7$`wya&K;7b#!TCVQe5KAVG6<a&K;7b#!TCVQe5pZf9w3Wo2?{Zf5`)EENP51Qh}R6#x|k6#@VN02Krko&yz>0!$Sc6$BCh04TLD{UFZ`1OQUV85>}2GX_L}OpZl}T5Jl`h5DCXLC_^~E~>EDrRz4CVh}dj_?LB?OfhH=8UM1_q~k15)7w1<axfw3HvllzJ#r8Kps=X)7z?ahR79YCBB%HTn%5+keg!o>#WulA{Dg)2tS9|S2mnGez@vi`$nC1+P6h--9)PbivpEJ802R6d70d!G70Lw_1QiAq1{^l50w@etX=7y|L1b-da$#&35khilXJKS%WpW@#Wo~71VQc^Z04TLD{UDv`1OP7p000NPEfE?4QT{ys2HSqmxlXt{xwksja&|a>2jdE;-_d5N!R`>3T>5<ha$4p2@&q*i5Hdj^A_4#aFd!lV0|Fod06`!?QYIk)0stZc0ssaCBOm|*A^-v*2Ot6v0U!W^2?PQF03ZS&A_O2HARqvuAVOjQ1R?+u0u%rT0s<l;03`<k04i_-fMQTY3jiP>0&xN&L?8ermXQJ+6$}|Q1ONsZ7!7S{Y-x01a(Q2JVQy;`3>L}-7VHKV^aUo?jRGd3ivk=rtO6(uR%v5pAVFkpX>ws~7!g8pX=h<%X=QRCM`dnha$#%$87vh96$BLm02Kfg1Qh}R000#P6@&#9zyeGa92Eo-001bpFa2Ob0JQ`F0tOZwU`=(Px^ok$Fw4OQ;C%w)4Z;h6`Ol=-goJ<u3-kARDOrfkQNW>4RbtmjBA?U700t4NLW$w+ybMrPa*#JB{n)F5;Kmywzk-GN`@EEl%@4U?oO4nc09;Anf(hi~DdFP+FzykG!eD9GBUQG3ybwshj1O=bIx|?t#lT=&84HL#!8kAkP&>i6JvWEhYyuSk6~F=&(gG|M$^{h!6$TXs95$o^C=pdtAWmU+c_1iKWprtBWn?KB3{_JgL1}UzMsIRsWdHyGD77#BAfN#S04@Lk00+D+5h?*sJpKmzf6uv2_&d3`I@WS_IDZG@3aH=FW~ssM5SLu~eF1V><@xdiHxMC1Wi?Y%5fLR}V>cLKAvsbKBNjFm2BQ%>5im11P!Tl|6H_rjH4zg$B{36G6=4xm5F;uTSTPeZGs7@3F)<=DFk=%D8DkM3BU2I;9Tqb)5i=1J6f>~~5(Z*s!v!%l!-WH8W-$T)049o=0vr_#88rj|1{oL+ZE0+2bYXINUt)4$ZewT_3>L}-7T5?DrUoX&l>#Q@k^&qyqyi`rRZ}2NVRm^SC{kr~X>w&`DHse@Qy@WUav(-;a${ux87vh96$BLm02Kfg1Qh}R000#P6@UR1PD~XL6$BCh04TLD{UAW~005u@0096500RI800saD0000000006PyzsBka!FM6#x~W0u{0XEEUQH6$BLq6$Ts_2nPTF0000002l}W000000000M04N9t000000000O2mk;80000000000D77#BAdrCs00RI30Lldc16v4$K>;8jBPJ{aIt{`aDRH215{WEHlq@)*1zRTw%9LeVpcF|01`Ec-z(ybdf+-~gNL+#<1%zyEqGS<PMgh{K3N%Q;AZZDbLLkN=+B!gs5+oRgAxttW69P~Hh7>@OEtrM@5vb-_vIIjmVuO|_5CCm~Dp42_5NJ0=0SrupfJI{r04DmL0vr_t88rj|1{oL)b8ul}WnX4&X=iA3a}^90$^{m_1QwnICdiutCfu3=92f`(000000000O2mk;80000002Tl!2nPTF0000002l}W000000000087vh96$BLm02Kfg1Qh}R000#P72qKi_y|lDoCFmF5&!@wwJ-f(a0itU05)HFNs!u2I?QoZ;hce6(I(8$r|?RvgBJ+fjQs2bOpWb_klNj_)`EU)fbZ=|ZbyZWu@Ra16EUPEy*8+h?G;mte4?%&N)xiBOfqHwV*q9VF_u`RKI<hD3rV||4n~6ZtdJ7U9lYS_CZ~Uv1r2fQN#30(tA8kk8!Dz3m&}{S65S}5-MC3<p%{>{4r55$lqfSHXNcXtIv6AaJ3Ya6_1cJo7^$>cs0#Mu6R^#*FWMwa;}U#TmCC4kq?1xF3i!+~97Ek6x3u|LSH9Ifhr+yC{B0R~J3dWOoesoD9o4*A@&wZ7;}J}(t#TUZEoQO>XZvH73`l?U2f7{GuM<q77OCkWkN6N*wU(CgT2t{UKDSX4aZO{34e{Js;2iz+o;h}6J*9X!{=A+I>n8@>gh3o>S2WlligPpQ1_mm1Y+ShFO#}<s$#<lNljZO*juEPcV$4#ju)ufi660!G!Qc>WG#*BUKscT+`am7SVeZq#aX3Qw4krb>ySuE_LFJQ5pmb4bOd$aZ?9S_SnU08QgmlmVz66~rjTo#=oMLc%txfsYUZz=9)1fpE=Z8A10rE0k)~dOditP(3bRJRqs)%z7hIeMdQHdO?zIV`VjhRhzxKNW#-OL=21Az^wAu#^o&jAr6)IJQxiOr?9Oa<MX{%YAcgbX7wwvC`Dpc`I^lM2?7+s9RuPYzYRCGs${VPZvM^|cbV2zg&(qyVUz7wD=L02RCj70?DO70Lw_1QiAq1{@d&5;6b)000007zh9W00000000&MC<qcV00000000;W00000000000000ewJ-f(o&fO-00RI30Lldc30nXH2pB-1fPn)D5-3=}pn-!25F$vJK%s(#01+-=$e>{ZhYlV-fCwRC1c?#>Oq@WGLPdlXELuc};Nk_07&2zisA1y<jtDt)MA*Tj$B7s|egFv~WC)QWMvfp!qGSn^CQhC}i6UhRl`2-QFbRMqLKZC`wshbUp^KL<V8VzQ69LBwGG(@yIfEvRni*>%kYUrt%^Nro;>bBOhfW<kcktxVv*!$-K7RfL02D~jAVP%<9YT~Sf}%xa7&UVAC_(^8kt9u`RLRmMOqn!o;?&6#iBF(Zgc3E1R4G$RBAh}MA(g6CtXjEx1uK@US+pe8ieT&330%2!b=k$MR}w1`eEli`ESRtx!-y4IW$ai4WQCF`8?cO7vnqL<sZd7a8MGPEq>0HijaoHp53XU0kWHJfZQQ!4DB&oJw=>_s*$O8v!Z>o}e3|2LE&{rA>eihiz^>gVcXr;vizjbh1bS}jMX)C-SA=``@)6PZVIRMKqWt{>7?7F3fe{E6G<XnULWK(>IAYik0000czOe!v6$BYI1ONsZ7!q@EVPj=qW^8F^Xmnp_Z*_8W6$}>21s1+37U~ct?x_MM)Sv<!7zh$F00000000;W0000000000762#+5;6b)000007zh9W000000000PEENP51Qh}R6#x|k6#@VN02KrkumKfZOcfRt1QGxMD77#BAXqH{0MG&e1ONsA0RRI42LK5G1po*D3IG5A000002~Yw6V~jDz7~~ZI75V}do&qct$^{h!6$TXs92f`+000000000O2mk;80000002Tl!2nqlI0000002l}W0000000000001bpFa03BH3R_Ciy$J9DQOM=e|)Mc;s8WrE>DrFoMlVT@-l>~5F*JwruIa3?*}C)2+E;yU~fZkd1neEVdrAR+^|FGA!$kTaXjVbq~L<kYorWfv2{{3NHaF<-t{K6!mHgAYX;g#0^HCHjA(x-!&<EtiJ*`-nby(8kyGm<Wl~zw!hr-T&N%@A04BP(0vr_t88rj|1{oL&b8ul}WnXt;Y-@8B3>L}-7RUt_m;@%Av;rogvjQ9#2nqlI0000002l}W000000000M04N9w000000000O2mk;800000000>*6$BLo6#@Vi02Krk0ssI26$BNq1{Lf9OciDo1QGxMD77#BV3+`90|26i5fhN`H3I-ZKzxN*006bX1{<tR0e?_I&dWy0PfWw=Way}-c2Jc12>4}Gj6C$rEZme{DG3uH0}-*EmD@$@lxx||I(h*yPyzsBj4{R-V~jDz7-Nhv#u#IaF~%5Uj4{R-V~jDz7~~ZI6}|)&j07wd$^{h!6$TXs92f{g000000000O2mk;80000002T%)2txn>0000002l}W0000000000001bpFa2O20Obb&MiD$WP{k%4uwsL5t64PFH>5Gr15lW$dA;RY+>|EmsM8*dvdO&@64y`wDCNVIQk|%^O&D=+vSiCF04e}101>@N&|5KG4N-3r{QCCu^!hz1739@7*HG)#76hzW#FZ3$7E{XdDv!(@2=(LqsL<7*`UlzTjOf=NjXy$R)H0|aO6f$}L!`qE>(Z~dy!vH$El~&cTP2A#ig~rDey<XZA0P@;A8M~Ry*|NV6}w=7Qv5u2-pG%aH{f+-huj01nMn~*PEbXp=0dCvW)3sYhduR19Qjs&5CMGSyqs5qqRH!y1LmprD)S3}BQOO50GBd=*-M=PCaAyy92Eo^H3R?#85j|BaA9L*Uw2__Yja;{Z*_8W6$}>21s0AB7MKSnoVx-hkhuaJ7zjfE00000000;W000000000076vE?LjV8(000007zh9W000000000PEENP51Qh}R6#x|k6#@VN02Krkyiygs6-*WE0~G`k001bpFa2Pi6QwEuCUJ=+&{nAq$W>w&^Ikq+A79|PF_)Tb+fqs*`4@<g%onK<fieaZ8;}5f0DA!cnlE<b<><a7UsAW%cqS=tr~7hm%gu)N$0kb~cyS@@pMgFfP)I`*MF4x|0r7lvCZZ?;nDG!)!s%#$L}1wLHf^S_jEH9hg!36`z%cHnG8_-a2gwKYfMIgmw3$Mk4Uo>~!{K<qFp2iZE}0V&6VQZV%X)6pq=}+PWlTby5>SSS2Li+9-=@tJMUfOy6ahr_R{v73^1ozL@K@<h@Y-^$bAR>F{w#7a<t|Mb;eN52Xnz)N%Avh4m64222PkyGMnpOr)()c5-(Y2o4)((#+*7na38MWm3tVq;TaaDkzot~35#jN9&kTd}G4Y`6bC=*;8=2oEw>saZNh2lP$Je~;98XCJk6qr9tWme1^4heSp!;I(e$AxW7Hl_H$<<g{b5-t2`_`<ymz=h2^3w_MHSLdCDiFH(E=6~J?rbXARI^L7OMHb_m)FNK|J;-fuDA-fO_N4Wsi10RVFj5K)dl8?a!tPOwU-jzgc4+ZjA5_Pk5xxz5+cb6bu0pz5U7OH(Ey21sZh3QGqDBtR=2HhW?fcz)*4&w>X-l2J*fX`uj1D2Pn|L$s$;HoDsy_r7|K@JZJIQdz4D&@td0XBD`eKBb~`D(pRAG}YqLUy(GvSD^{=&KE7Z7WPp4{qt=-C909T)To2#NdfwQO~No$Bil9D7@)`0~{DF)?v5g-&rlFTrwL`2#otpUKHMFt9)yBGx?#(~MRfu9m+2%tv(je<zB87Hg!f<;(v;gqm1#Q5DE`4CZp!>%z=AH7}fD%isZ@{i92_i?jLJP28>9kGlx)-sdf^RupLS<_8W77jmJDHS4lxkwM;FC>7_=y?H^!DuiG6g=Ol*=+)y(Cl3{@E2_FM<#qY=?BwdemENnpmoID`<922l_cTO-56WozpCF*wi2WF0^$q(X`^V_RQ72b3hw1TIr{JfxkBRHE@s;T@p({us~=}GM8MfXRoptDz`tx;#Y+MOLk#N*&D#@_%@qO)O}Fx^X0#wuZ8Tx!?u|=D<et;R<fJL_1F@t?TL>sG?W|0RLjGL-*C@eGc+%le{tvRD6+dR6H=7|Mf~00wLk?_9x)rCfrKI}!cvpT$V90Qv8^LEfn#UdDXBBFxn#U1zT*7ezTf`TO4;-+=E+<m|UrM~ck*Aley{Gjt7J<lOwkaGu5NeQ`ZKlGBu!WA6(uJ20P7L>5SDdq0$>pz<diPVnFj6so#E7S4hMfLXL)r)c62SOmovdYso@iNC5XR`ne>S|ummNMIu|KsKC$(sdIYvOpCl#R+wD0sXkl?ilHJs?hX8Q4nro?#uOlXVA<tS~${5ZDP#wrYDdMHT1X-q*Q3gWyd*GW-bCecqkhu|S&5V-J6_l!ZFGlTiCR%U-3tA9SDjw6P_wg=?q$@99qWsw#r8<1Dp)~`jaBXMAA)z))(Wi%)zha5$fR0+S<CzEp$k8H^GM!5fHxJTM7bGc|Cw+3MV&#<Kjfw6th>p!vuI;Iyb7FMYGbPX1Va(9}IBou;h5lZj!dQFUSi~6|MpZwY)KRkm-srzX^1)1nhlxQH--{Tyh%Cv;=!B4ybl902k1$5{5TC>s+l4$bn@}OmRO;{|1l=L{ONwV90z|olZXB)t_8>9kZp@Bq>j3kG`>XE426pvaA&|bel1N0*B7j85xZ=4NrLNQxNWg_%Y7N#rM$icQQR(hDEp#L$K>cf&M6#y055EbqaEEUQP6$BLq6$TtOE+`07Yk6#8Vjwga2tjafY;12JF#rGnD77#BU@Ze_Apmax004i|wN7<q+BNF!*0tx3Psi@a;Fx~8X?we_vuk$Kc|kM{Uu{aIou{iiV=u2R+j|^y?QzQEoo?N!?ry~E(pg<@d#jrxT5ExQXP0tnc@;M8*2?9cpD|iCZQVFsb(`+pxhARiwsY#Mov!A_7h?bb08Rq~{QnRG1N{GL0|os5|L`^g1N{G>0Rsj6|NrwK0|Wg3C<6oh{}uxS{Qrgl0|Wg3q5%T~{6ByJ0|Wg3Farbp|8fHZ{QpJ+1N{GX0|)&7&;R`YSOWw6{~QAa{QnOC2!8+p0s{mE1_uZU3JVMk4i69!5)%{^78e)*85$cL2ps?(3IHD<AtECrB_;$XC@Cr{EG;eqFEBAOA~OUuH8wXmIXXK$Jw87`D?vg-L`4)p87n0RMn_0VN=r;lIZjUoPy!oKQd3k_7*<zUIwDyR5n5YZC{kSoJYHX5VPYC%WD;d&GiPWFQfX>yY;A6DaB3tAL=ACra~NJMbae(Cb{KbfJREj;B0FPxdwg<6AANp*fPsPvgM=@IhJ%NQi3f^KXIG1ijgF6yiCmGAQ+$(@m0uDYL{lYTD?^klbe1|#B_<b5mlhY20+^X?j80lWnluw|o1AbDnL}fKomHNnpfOM`9iO41qd26c0wbcPIS;2zJ}{_>R}6ZOkf~-&YNs0}s#F>rcB>N^gfDTdF{`b3u9UAcu$MluvTm0tRTB~;9J5|zkC3#Y89TK*XeLLtevP+zI*bp9d1^vOA855nw73GWu(p*+pSe2*UX!|rXRZ)3sbIT4O$!GNybVXRE+f5~Ef~H}3k)fIzZ)YbmwUi0DK}aZF>o9xdwg-Mslj5ap`vhAV8X+NkB~sb#m1FMbF7y(abkEIwMUKvLrX@eL$AkkRH>mWwa7^j5;2dEa=AuyxyhM`O)#aTMF|wMPXv^I5z0;^1T<I6A<UY+%@Pw4U>44nw^$Q2WOUEJj!Z%rB+$`2Xi-eRi6ylK26Q3Pq)-;qU^scFXFUQlum^iiIvTcWjcSNodSBFkH=&}{cOzp_EjW`MpIE=HW0y;>tx+~qAks~js7;eH5Li9dK-VDH0A{2XeA&IWCVz3RTnJZ07pWcE8F|0K0|h0ujE+-OVlqa1Q!OCV+uYr@WM4SmuxL_7P~YH^ZPAUt9?`xmbB;825I!R`M0O#e8K`#y;nNxgC*oK_3rfEzM^!j?B8lK@Y)V={Os(T9UoIV=jzU8n0OYf)<xxjB=8Ug+uIEzd!D3#OkB~D42Sz!sX9f(IIj@bGJ0~*fusnLSxRid6kgPbQnt48V4|qI!>Ret3U9Xk2g*{Ivem23m((2G|>+J0)2p17PB((1D@bU6c^P$of^pq$mboC*<zOsUBmW!y;G5`Q3uIU0C6$}|Q1ONsZI0$WNb8~5LZgT`|X>)W0X>?_66$}>24Hl4O7NjR8=+^=!ioyaMHZCX#Q)_u_VPYUO7zjadZ)|LDATa<LEENP51Qh}R6#x|k6#@VN02KrkumTm90!$T56$BCh04TLD{UDeX1OS4GA1r`NGY0@d<MiQ8WsD1MQzcBC0t3UGCC}wmSSqez4=$qFY;BWho0ttQm#23b#*@Tta;D-+!W4(2g(Zpw4h}?8jk%sKZrCm23R|PF0m2o-1rs-posXf6qmC~YcbRn;0`cRq<>7$q&2u-RF=S1EJ>X19cOnh9jRB|t6#x~|0Ttu{EEUQH6$BLq6$Ts_1ll|R02l-|2lf9J;shuJ+B^UN7z8#4_5T0>04TLD{UB*c0{}Sy000yK1rb0L01ZKeK!^kk8~_6m6aWhWAb<ft8~_6g8~_UeAOc7P8~_hd0t7Nig2)sA4MBh;AOQdd4?rLc08tnR03eC7FpYCO02E0R6#-EdSr-;%+QxN};C&GoiX(ZH0000c=I;U=6#^MG1ONsZ7z$)%VRUF;WMOn=6$}>21r~$_7W@PzpzQ)C>goa<7zElp000;SHV5_p7UBdb1ll|R02l-|2lf8|87vh96$BLm02Kfg1Qh}R000#P6?PSf0ZbJJ6$BCh04TLD{UAHp0RR#J002`|Q!OztHy|iOZ)|mKVrgM12mk;8L2z$uY;Pbj6#xJLQ&dwRH8CwrASgp<VQg$=Zf9k3DHQ+}Y!!Yi70Lw_1QiAq1{^l31t=0zR8uW6FgGA5LvL(#ZDMI*DHsSraBpmEZy+%M001bpFa01ku>k-q0000K00RLq00RLU00WVV0RRI5EC2%uQUC*qE&u}o8UO<U5C8=MKuiDw0T2KJ1rUUyY9^}m0vr_#88rj|1{oL)WMyG=XkTq<b8~5LZWRm`$^{nc0v3(~Cc^RpCY<pC95$*2C=ydtQ!OztHy|iOZ)|mKVrgM17zjadZ)|LDATa<LEENP51Qh}R6#x|k6#@VN02Klm{2Da?7zT4=Xk~3-6(#@`3?%{>1a4t%WhDSi8~^|S6$B*$7zBA`VR8T!1SJ9(1!r<^b#MR`3?%{>26J>_baitj08AVJ000#NB?1@<Vrg=8XkTPubY&$BOl$xE05t*_3u0+<bZB32Y+++%6$Ahk3?%{>1#@&^bY&#~OdJ3J02K@+0vH5zZ+C7b08AVJ000#XB?1@*XJu|=WpX6|OdJ3J05t*_3Sn$*VP9=wbY*fC0u=^KJOBUyH3S$0Y-w|J6#@Vi3?%{>1Zi|-Z6yFq8~^|S6$~W;7!!4FWMy(^a$#h3VRU6*ZDDF}awPyv8~^|S6$~W;7!GH0VPth-bY)*{VQO!3B>+qu0000L3?%{>4Q**`X>?(7d0%p2Zfhj~OdJ3J02K@+0vHc%X>4h9VRCt2Vsc?_V`wD+OdJ3J02Kr!0vHW*aA9L*UuJA+XJ~YD02Kr!0vHl=aA9L*UuJA+XJ~X^Xm53La{v_tB?1@=b8ul}WnXt;Y-@7>6$B*$7!h-DVPj=qcVTR6b6;q0b#ik6H3Aq2ZE16JX>V?G6#^9oOgsPp05t>{1Z-(@bQJ;s6$~W;7zAl_Wo;z@OdJ3J02Kly0vHNpWnpw^Uu0o)WhD$uYybcN6$~W;7!71)VRUF;ZE16JX>V>N08AVJ000)s1sNO}{1hhKLL3zg88rj|1{oLxZeeX@6$}>21s0Tj7Suu}jyom@95w(b4_S6^Zf<2DL@pphVP|Y*7!*NlVQwHoE+A8AWpH6~WFSg%E+7CIEENP51Qh}R6#x|k6#@VN049b@92Eo^H3R?#85jh4Wnpp^3>L}-7JvvA>;@*%MJAd=92f}M2LJ#70000O2&V@C0000002VeV2-yb!0000002m0T2LJ#70000087vh96$BLm02Kfg1Qh}R001VKQydip88rj|1{oLyXL4_Ka1{&|$^{nw2Nuu<Cge>f$V?m<2owMS0000002l}X000000000MHYf-b000000000O2mt^900000000>*6$BLo6#@Vi02Krk0ssI2CX`ql6$}|Q1ONsZ7zT56VRUtK6$}>21s0S87Ptc@=v5}%R2(({C<jtybZK&BWEciPV{~bDWdIp06$BLo6#@Vi02Krk0ssI2CX8<!6#^MG1ONsZ7z$!(a&%~4WMOn=6$}>21s0wi7T^^o+-D}NSsWMyhztM#7zEwA|Nj;MC<KTM000;S-Mat(02wS51Qi4o0ss{N6$BLm0000c%6l9Y3>h^900tQ-3u0+<bZB32Y+++%1#@&^bY&F`7Rm({#0?gZ3ntoiCc1DOHUKCLS7mc_AV+0#ZDDSC7y>~|02wS51Qi4o0ss{N6$BLm0000co}mF86$}|Q1ONsZC<|g~a&%~4aBN{?Wdw9@cWxC77Rm({_B$4&BqsQo0Veu<95w(b2U%}!Yh`kC7zjaPX>Mn9Z*Bk?EENP51Qh}R6#x|k6#@VN049W`0UQ+!88rj|1{oLzXJu|=WpWh^7Rm({jsh040w$!R0VdL-0UR~}C<IMmY-Jb*MrCbbY-Io$EENP51Qh}R6#x|k6#@VN04A2y0UQ+!88rj|1{pXCVQg(-Uu|J@WpV^;X>)W0X>?_66$}>A3Kq_57WyqF;Ku<boTdRBHUKCaS#Nc2XLEFKcWxk4bYXO5AXRQ@c4cyNX>@rQ000>*6$BLo6#@Vi02Krk0ssI2Ccfwa92E>1H3R?#85k3FZe(S0XL4a=bzyX6Uu|J(Z*mn37Rm({vL+VN5+;!10VeF#0US0gC<j_^Z)|U8c^C>oV`Fc1ZggpGX8;*26$BLo6#@Vi02Krk0ssI2ChCX+92E>1H3R?#85j;{a$#h3VRU6*ZDDF}auo~~$^{n0Qx@zVCY*r+CjRLG95!|+DpqB5WpZh5VRCsOO=V<hV`*+>J0MVXVr*$+AVFkpX>MtAbaG*IX>V>AC_!a%Z*F0AbZKK@Y#=5eL34C+Z*F0AbZKK@Y#>E$XK8L_WpZh5X8;*26$BLo6#@Vi02Krk0ssI2CYF%`92E>1H3R?#85j+1X>4h9VRCt2a$#<36$}>21s3cE7W4%s){O!tqKg6?Hmm|D3|47lWgtOhZE130Y#0$ja%pE_WNBq`AV+0xWpZI`02wS51Qi4o0ss{N6$BLm0000cikSi&6$}|Q1ONsZ7!Pe}Y-x01a(Q23a$#;`XcY_=$^{nK2o|OWCd8EjCghR=95$o^C=pdtAWmU+c_1iKWprtBWn?KB3{_JgL1}UzMsIRsWdIp06$BLo6#@Vi02Krk0ssI2Ci<QN92Eo^H3R?#85j+7aA9L*UuJA+XJ~YD6$}>21s1*p7M=tq$eRKt+?oO$7zhUd00000000;W0000000000762#+2LJ#7000007zh9W000000000PEENP51Qh}R6#x|k6#@VN04Bb%0vr_t88rj|1{oL<b8ul}WnX4&X=iA3UubW2a&r|77Rm({z9|;!5GL-a0w&a;0vs3!5;6b)000007zh9W00000000&MC<qcV00000000;W000000000002wS51Qi4o0ss{N6$BLm0000cy0-!x6$BYI1ONsZ7z=Z7VPj=qcVTR6a}^90$^{n41s0eDCY-baCZe+f92f`+000000000O2mk;80000002Tl!2nqlI0000002l}W000000000087vh96$BLm02Kfg1Qh}R001Vazycf<1Q|6100tQt5p!^1V`X1=VQg!2UubW2a&r|77Rm({jtmx<2PT}m0w$2T0vs3!LjV8(000007zh9W00000000&SC<sFU00000000;W000000000002wS51Qi4o0ss{N6$BLm0000cuIU0C6$}|Q1ONsZI0$WNb8~5LZgT`|X>)W0X>?_66$}>24Hl4O7NjR8=+^=!ioyaMHZCX#Q)_u_VPYUO7zjadZ)|LDATa<LEENP51Qh}R6#x|k6#@VN04C<|0vr_r88rj|1{oL%WMyG=XkTPubY&F`7Rm({gasD-1SX*E0w(I}0vs3w+B^UN7z8#4_5T*)1SkaBJOBU~1U3it{{R^*6$BLo6#@Vi02Krk0ssI2CaUxT92E>1H3R?#85j*@Wnpw^Uu|h~b7^mG6$}>21s3W87LEfZ!tw$pobdu2HmU_E5>r%DEio`RASgp`Y;|p7X<;cC2tjafY;12JF#s7X6$BLo6#@Vi02Krk0ssI27UY=)7Rm)C2o}cH0u%rM85|f4K~hprS2}ZJXk~3-7?=$&FE1}ID`!PPK|xVLK|w)5K|xDFL3cqyXF*0mL1#fjXF)+hK|w-7K|xJHK|w)7K|x7DK|w)6K|x7DK|w)6K|w)5LQO$IK|@JGK|w)KK~X_LL1#rlK}2IgcR@jKK}|tHLT5xlK}2&wXF)-0K}kVDMrT7oK|yOlXF)+YK}A79Om{*-K|?`7Q9(gbK}A79L1#iiK}1PGK|w)cK|w)5L}x)kK|w`9K|w(=P<AgbcR@ixQB_evK|w)DK|w)5K~X_LK|w)5K|w)8Q9(gLM^bEASW<6TGBI#tGfr?~GBiO!K}1a}FE1}bP<k&fcR@ixQD;#>K|w)CK|w)5K~X_LK|w)5K|w)7XF)+hM^bEASW<6TGE!({M^!;VO>#IdFE~L#K|w)IP<k&fcR@ixQFm2AK|w)DK|w)5K~X_LK|w-6K|w)BK|w)5L1#fhK|*w4SWbFjS2=ZKcR@ixK|ymbP%kf9P<k&fcR@ixQB_evK|w)BK|w)5K~X_LK|w)5K|w)6K|w)5M{F=!S1>_AK|w)QP<k&fG<rBMFE~L#K}ADFK|w)5Ls3CNK|w`9K|w)5K|w)5LPbGAK|*?CM@DK|Ryi|SS9o$`F*$EyGgf(HcR@ixPDD5_FE~L#K|w-MK|w)5O=>tVFE~L#K}ADFK|w)5LU%zyK|w`9K|w)5K|w)5L32StK|*?CM@DK|Ryi|SS9o$`cR@ujW-l*6K|w)5RY5^PK}0w&P%ke*K|w-6XhA_iK|yOlK|w)6K|w)5K|w)5K|xbNK|w)dGeT%tGFE0<ZFq26GH5VZGH7;WSW0?9K|(<<W-l*6K|w)5RY5^PK|wSxW-l*6K|w-6XhA_iK|yChK|w)6K|w)5K|w)5K|xMIK|w)dGeT%tGFE0<ZFq26GH5VkcR@ixNNPDRFE~L#K|w-MK|w)5NJKd=FE~L#K}AMIK|w)5LqS17K|w`9K|w)5K|w)5K{!D{K|*w4S9ow_M@D&NRyk>7ZANujGG{?yP%|$tHbgluFE~L#K}AMIK|w)5LqS17K|w`9K|w)5K|w)5K`=o<K|*w4S9ow_M@D&NRylcDS2c1$K|w)nP%|$tIYc=xFE~L#K}AMIK|w)5LqS17K|w`9K|w)5K|w)5K{P=@K|*e0ZANQkS4J>eRylN8S8#S?XF)-EP%|$tcSI{MFE~L#K}AMIK|w)5Ls3CNK|w`9K|w)5K|w)5LQO$IK|(QNZBlGwZDe^_S5h%qSW;|QGBIdjZ8>>CP<lBpFE}tSb}uhMK|w-9LQz3MK|yChK|w)6K|w)5K}A79K|xtTK|w)7XF)+hM@DmES4L)HS4J>eSV=)aLwYMOFF9*3b}uhMK|w-7Qb9pMK|xtTK|w)6K|w)5K|w)5K|w`9K|w)bSW;|aQ9(gLK|)n8b}uh#P&Y3xcR@ixQB_evK|w)BK|w)5K~X_LK|w)5K|w)6XF)+hM`UbbZBlGwXF)+`P&Y3xH&iPxFE~L#K}AkSK|w)5Ls3CNK|w`9K|w)5XF)+hMNvUPK|xtTK|w)7cR@ixM@n#GZc=YpGeUD&S59m}OhhX$FE4K|b}uhMK|w-7Qb9pMK|xtTK|w)6K|w)5K|w)5K|w`9K|w)eM>#WMXF)+hK}1<Eb}uhQP%kepcR@ixQB_evK|w)BK|w)5K~X_LK|w)5K|w)6Q9(gLNKP<WSW;|3K|w)5P%kepdTK8(FE~L#K}AVLK|w)5Ls3CNK|w`9K|w)5K|w)5L2p4pK|*O^SV}NqMmcL)SW;|3K|w)5L0LgVK|x1BL0LgTK|w)5K|w)dP%kepcR@ixQB_evK|w)BK|w)5K~X_LK|w)5K|w)6XF)+hNKP<WSW+=#cR@i}P%kepHEJ&}FE~L#K}ADFK|w)5LT5oiK|w`9K|w)5K|w)5K~+IPK|*e0Z8<SxK|w)5PDC#+FE~L#K|w-MK|w)5O=>SMFE~L#K}ADFK|w)5LqS17K|w`9K|w)5K|w)5K~X_LK|(cJS4Me3K|w)5L1#fkK|x7DL1RHfK|w)5K|w)5K}JDAK|w)9K|xhPL1#fiXF*6oL3cqxK|xVLLP0@6K|w)5K}AMIK|w)5LU%zyK|w`9K|w)5K|w)5K~X_LK|*y|S21isK|w)5K~X_MK|w`9K|w)5K|w)5K|MVH7$0z9a&dKKbS-0Wa4lhSa&LDac4cyNX>V>II4&?QFd6(58~^|u000~S02}}S8~^|u000~S02}}S8~^|u000~S02}}S8~^|u000~S02}}S8~^|u000~S02}}S000pH08l|vF#"
        )

    return pq.read_table(io.BytesIO(_decoded))


def schema():
    return RelationSchema(
        name="$astronauts",
        columns=[
            FlatColumn(name="name", type=OrsoTypes.VARCHAR),
            FlatColumn(name="year", type=OrsoTypes.INTEGER),
            FlatColumn(name="group", type=OrsoTypes.INTEGER),
            FlatColumn(name="status", type=OrsoTypes.VARCHAR),
            FlatColumn(name="birth_date", type=OrsoTypes.DATE),
            FlatColumn(name="birth_place", type=OrsoTypes.STRUCT),
            FlatColumn(name="gender", type=OrsoTypes.VARCHAR),
            FlatColumn(name="alma_mater", type=OrsoTypes.ARRAY),
            FlatColumn(name="undergraduate_major", type=OrsoTypes.VARCHAR),
            FlatColumn(name="graduate_major", type=OrsoTypes.VARCHAR),
            FlatColumn(name="military_rank", type=OrsoTypes.VARCHAR),
            FlatColumn(name="military_branch", type=OrsoTypes.VARCHAR),
            FlatColumn(name="space_flights", type=OrsoTypes.INTEGER),
            FlatColumn(name="space_flight_hours", type=OrsoTypes.INTEGER),
            FlatColumn(name="space_walks", type=OrsoTypes.INTEGER),
            FlatColumn(name="space_walks_hours", type=OrsoTypes.INTEGER),
            FlatColumn(name="missions", type=OrsoTypes.ARRAY),
            FlatColumn(name="death_date", type=OrsoTypes.DATE),
            FlatColumn(name="death_mission", type=OrsoTypes.VARCHAR),
        ],
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from enum import Enum
from enum import auto
from typing import Generator
from typing import Optional

import pyarrow
from orso.tools import random_string

from opteryx.models import QueryProperties
from opteryx.models import QueryStatistics


class OperatorType(int, Enum):
    PRODUCER = auto()
    PASSTHRU = auto()
    BLOCKING = auto()
    _UNKNOWN = auto()


@dataclass
class BasePlanDataObject:
    operation: Optional[str] = None
    operator_type: OperatorType = OperatorType._UNKNOWN
    query_id: str = None
    identity: str = None

    def __post_init__(self):
        # Perform actions after initialization
        if self.identity is None:
            self.identity = random_string()
        if self.operation is None:
            self.operation = self.__class__.__name__.replace("DataObject", "Node")


class BasePlanNode:
    _producers = None
    operator_type = OperatorType._UNKNOWN

    def __init__(self, properties: QueryProperties, **parameters):
        """
        This is the base class for nodes in the execution plan.

        The initializer accepts a QueryStatistics node which is populated by different nodes
        differently to record what happened during the query execution.
        """
        self.properties = properties
        self.parameters = parameters
        self.statistics = QueryStatistics(properties.qid)
        self.execution_time = 0
        self.identity = random_string()
        self.do: Optional[BasePlanDataObject] = None

    def to_json(self) -> bytes:  # pragma: no cover
        import orjson

        from opteryx.utils import dataclass_to_dict

        return orjson.dumps(dataclass_to_dict(self.do))

    @classmethod
    def from_json(cls, json_obj: str) -> "BasePlanNode":  # pragma: no cover
        raise NotImplementedError()

    def set_producers(self, producers):
        self._producers = producers

    def config(self) -> str:
        return ""

    @property
    def name(self):  # pragma: no cover
        """
        Friendly Name of this node
        """
        return "no name"

    def execute(self) -> Generator[pyarrow.Table, None, None]:  # pragma: no cover
        pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The SQL Connector downloads data from remote servers and converts them
to pyarrow tables so they can be processed as per any other data source.
"""

import time
from decimal import Decimal
from typing import Any
from typing import Dict
from typing import Generator
from typing import Tuple

import pyarrow
from orso import DataFrame
from orso.schema import ConstantColumn
from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.tools import random_string
from orso.types import PYTHON_TO_ORSO_MAP

from opteryx.config import OPTERYX_DEBUG
from opteryx.connectors.base.base_connector import DEFAULT_MORSEL_SIZE
from opteryx.connectors.base.base_connector import INITIAL_CHUNK_SIZE
from opteryx.connectors.base.base_connector import MIN_CHUNK_SIZE
from opteryx.connectors.base.base_connector import BaseConnector
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import MissingDependencyError
from opteryx.exceptions import UnmetRequirementError
from opteryx.managers.expression import Node
from opteryx.managers.expression import NodeType
from opteryx.third_party.query_builder import Query


def _handle_operand(operand: Node, parameters: dict) -> Tuple[Any, dict]:
    if operand.node_type == NodeType.IDENTIFIER:
        return operand.source_column, parameters

    literal = operand.value
    if hasattr(literal, "item"):
        literal = literal.item()

    name = random_string(8)
    parameters[name] = literal
    return f":{name}", parameters


class SqlConnector(BaseConnector, PredicatePushable):
    __mode__ = "Sql"

    PUSHABLE_OPS: Dict[str, bool] = {
        "Eq": True,
        "NotEq": True,
        "Gt": True,
        "GtEq": True,
        "Lt": True,
        "LtEq": True,
        "Like": True,
        "NotLike": True,
        "IsTrue": True,
        "IsNotTrue": True,
        "IsFalse": True,
        "IsNotFalse": True,
        "IsNull": True,
        "IsNotNull": True,
    }

    OPS_XLAT: Dict[str, str] = {
        "Eq": "=",
        "NotEq": "!=",
        "Gt": ">",
        "GtEq": ">=",
        "Lt": "<",
        "LtEq": "<=",
        "Like": "LIKE",
        "NotLike": "NOT LIKE",
        "IsTrue": "IS TRUE",
        "IsNotTrue": "IS NOT TRUE",
        "IsFalse": "IS FALSE",
        "IsNotFalse": "IS NOT FALSE",
        "IsNull": "IS NULL",
        "IsNotNull": "IS NOT NULL",
    }

    def __init__(self, *args, connection: str = None, engine=None, **kwargs):
        BaseConnector.__init__(self, **kwargs)
        PredicatePushable.__init__(self, **kwargs)

        try:
            from sqlalchemy import MetaData
            from sqlalchemy import create_engine
            from sqlalchemy.pool import NullPool
        except ImportError as err:  # pragma: nocover
            raise MissingDependencyError(err.name) from err

        if engine is None and connection is None:  # pragma: no cover
            raise UnmetRequirementError(
                "SQL Connections require either a SQL Alchemy connection string in the 'connection' parameter, or a SQL Alchemy Engine in the 'engine' parameter."
            )

        # create the SqlAlchemy engine
        if engine is None:
            self._engine = create_engine(connection, poolclass=NullPool, echo=OPTERYX_DEBUG)
        else:
            self._engine = engine

        self.schema = None  # type: ignore
        self.metadata = MetaData()

    def can_push(self, operator: Node, types: set = None) -> bool:
        if super().can_push(operator, types):
            return True
        return operator.condition.node_type == NodeType.UNARY_OPERATOR

    def read_dataset(  # type:ignore
        self,
        *,
        columns: list = None,
        predicates: list = None,
        chunk_size: int = INITIAL_CHUNK_SIZE,  # type:ignore
    ) -> Generator[pyarrow.Table, None, None]:  # type:ignore
        from sqlalchemy.sql import text

        self.chunk_size = chunk_size
        result_schema = self.schema

        query_builder = Query().FROM(self.dataset)

        # Update the SQL and the target morsel schema if we've pushed a projection
        if columns:
            column_names = [col.schema_column.name for col in columns]
            query_builder.add("SELECT", *column_names)
            result_schema.columns = [  # type:ignore
                col
                for col in self.schema.columns  # type:ignore
                if col.name in column_names  # type:ignore
            ]
        else:
            query_builder.add("SELECT", "1")
            self.schema.columns = [ConstantColumn(name="1", value=1)]  # type:ignore

        # Update SQL if we've pushed predicates
        parameters: dict = {}
        for predicate in predicates:
            if predicate.node_type == NodeType.UNARY_OPERATOR:
                operand = predicate.centre.current_name
                operator = self.OPS_XLAT[predicate.value]

                query_builder.WHERE(f"{operand} {operator}")
            else:
                left_operand = predicate.left
                right_operand = predicate.right
                operator = self.OPS_XLAT[predicate.value]

                left_value, parameters = _handle_operand(left_operand, parameters)
                right_value, parameters = _handle_operand(right_operand, parameters)

                query_builder.WHERE(f"{left_value} {operator} {right_value}")

        at_least_once = False

        convert_time = 0.0

        with self._engine.connect() as conn:
            # DEBUG: log ("READ DATASET\n", str(query_builder))
            # DEBUG: log ("PARAMETERS\n", parameters)
            # Execution Options allows us to handle datasets larger than memory
            result = conn.execution_options(stream_results=True, max_row_buffer=25000).execute(
                text(str(query_builder)), parameters=parameters
            )

            while True:
                t = time.monotonic_ns()
                batch_rows = result.fetchmany(self.chunk_size)
                self.statistics.time_waiting_sql = time.monotonic_ns() - t
                if not batch_rows:
                    break

                # convert the SqlAlchemy Results to Arrow using Orso
                b = time.monotonic_ns()
                morsel = DataFrame(schema=result_schema, rows=batch_rows).arrow()
                convert_time += time.monotonic_ns() - b
                yield morsel
                at_least_once = True

                # Dynamically adjust chunk size based on the data size, we start by downloading
                # 500 records to get an idea of the row size, assuming these 500 are
                # representative, we work out how many rows fit into 16Mb (check setting).
                # Don't keep recalculating, this is not a cheap operation and it's predicting
                # the future so isn't going to ever be 100% correct
                if self.chunk_size == INITIAL_CHUNK_SIZE and morsel.nbytes > 0:
                    self.chunk_size = int(len(morsel) // (morsel.nbytes / DEFAULT_MORSEL_SIZE)) + 1
                    self.chunk_size = (self.chunk_size // MIN_CHUNK_SIZE) * MIN_CHUNK_SIZE
                    self.chunk_size = max(self.chunk_size, MIN_CHUNK_SIZE)
                    self.chunk_size = min(self.chunk_size, 1000000)  # cap at 1 million
                    # DEBUG: log (f"CHANGING CHUNK SIZE TO {self.chunk_size} was {INITIAL_CHUNK_SIZE}.")

        if not at_least_once:
            yield DataFrame(schema=result_schema).arrow()

        # DEBUG: log (f"time spent converting: {convert_time/1e9}s")

    def get_dataset_schema(self) -> RelationSchema:
        from sqlalchemy import Table

        if self.schema:
            return self.schema

        # Try to read the schema from the metastore
        self.schema = self.read_schema_from_metastore()
        if self.schema:
            return self.schema

        # get the schema from the dataset
        # DEBUG: log ("GET SQL SCHEMA:", self.dataset)
        try:
            table = Table(self.dataset, self.metadata, autoload_with=self._engine)

            self.schema = RelationSchema(
                name=table.name,
                columns=[
                    FlatColumn(
                        name=column.name,
                        type=PYTHON_TO_ORSO_MAP[column.type.python_type],
                        precision=(
                            None if column.type.python_type != Decimal else column.type.precision  # type:ignore
                        ),
                        scale=(
                            None if column.type.python_type != Decimal else column.type.scale  # type:ignore
                        ),
                        nullable=column.nullable,
                    )
                    for column in table.columns
                ],
            )
        except Exception as err:
            if not err:
                pass
            # Fall back to getting the schema from the first row, this is the column names, and where
            # possible, column types.
            # DEBUG: log (f"APPROXIMATING SCHEMA OF {self.dataset} BECAUSE OF {type(err).__name__}({err})")
            from sqlalchemy.sql import text

            with self._engine.connect() as conn:
                query = Query().SELECT("*").FROM(self.dataset).LIMIT("1")
                # DEBUG: log ("READ ROW\n", str(query))
                row = conn.execute(text(str(query))).fetchone()._asdict()
                # DEBUG: log ("ROW:", row)
                self.schema = RelationSchema(
                    name=self.dataset,
                    columns=[
                        FlatColumn(
                            name=column,
                            type=0 if value is None else PYTHON_TO_ORSO_MAP[type(value)],
                        )
                        for column, value in row.items()
                    ],
                )

        return self.schema



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Callable
from typing import List
from typing import Optional

ONE_HOUR = datetime.timedelta(hours=1)


class BasePartitionScheme:
    """Implement a partition scheme"""

    def hourly_timestamps(self, start_time: datetime.datetime, end_time: datetime.datetime):
        """
        Create a generator of timestamps one hour apart between two datetimes.
        """

        current_time = start_time.replace(minute=0, second=0, microsecond=0)
        while current_time <= end_time:
            yield current_time
            current_time += ONE_HOUR

    def get_blobs_in_partition(
        self,
        *,
        blob_list_getter: Callable,
        prefix: str,
        start_date: Optional[datetime.datetime],
        end_date: Optional[datetime.datetime],
    ) -> List[str]:
        """filter the blobs acording to the chosen scheme"""
        raise NotImplementedError()



# cython: language_level=3
# Public domain
#
# I hereby place my work 'mbleven' into the public domain.
# All my copyrights, including all related and neighbouring
# rights, are abandoned.
#
# 2018 Fujimoto Seiji <fujimoto@ceptord.net>
# https://github.com/fujimotos/mbleven/blob/master/LICENSE

"""An implementation of mbleven algorithm"""

# This has been modified for Opteryx for performance.
# This is roughly twice as fast as the original implementation.
#
# This file maintains the Public Domain Licence, this does not change the licence
# of any other files in this project.

#
# Constants

REPLACE: int = 1
INSERT: int = 2
DELETE: int = 4

# fmt:off
MATRIX = (
    (
        (INSERT, DELETE, ),
        (DELETE, INSERT, ),
        (REPLACE, REPLACE, ),
    ),
    (
        (DELETE, REPLACE, ),
        (REPLACE, DELETE, ),
    ),
    (
        (DELETE, DELETE, ),
    )
)
# fmt:on

#
# Library API


def compare(str1: str, str2: str) -> int:
    len1, len2 = len(str1), len(str2)

    if len1 < len2:
        len1, len2 = len2, len1
        str1, str2 = str2, str1

    if len1 - len2 > 2:
        return -1

    models = MATRIX[len1 - len2]

    res = 3
    for model in models:
        cost = check_model(str1, str2, len1, len2, model)
        if cost < res:
            res = cost

    if res == 3:
        res = -1

    return res


def check_model(str1: str, str2: str, len1: int, len2: int, model) -> int:
    """Check if the model can transform str1 into str2"""

    cost, idx1, idx2 = 0, 0, 0
    while idx1 < len1 and idx2 < len2:
        if str1[idx1].lower() != str2[idx2].lower():
            cost += 1
            if cost > 2:
                return cost

            option = model[cost - 1]
            if option == DELETE:
                idx1 += 1
            elif option == INSERT:
                idx2 += 1
            elif option == REPLACE:
                idx1 += 1
                idx2 += 1
        else:
            idx1 += 1
            idx2 += 1

    return cost + (len1 - idx1) + (len2 - idx2)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Expressions describe a calculation or evaluation of some sort.

It is defined as an expression tree of binary and unary operators, and functions.

Expressions are evaluated against an entire morsel at a time.
"""

from enum import Enum
from typing import Callable
from typing import Dict
from typing import Optional

import numpy
import pyarrow
from orso.tools import random_string
from orso.types import OrsoTypes
from pyarrow import Table

from opteryx.exceptions import ColumnReferencedBeforeEvaluationError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.functions import FUNCTIONS
from opteryx.managers.expression.binary_operators import binary_operations
from opteryx.managers.expression.ops import filter_operations
from opteryx.managers.expression.unary_operations import UNARY_OPERATIONS
from opteryx.models import LogicalColumn
from opteryx.models import Node

from .formatter import ExpressionColumn  # this is used
from .formatter import format_expression

# These are bit-masks
LOGICAL_TYPE: int = int("00010000", 2)
INTERNAL_TYPE: int = int("00100000", 2)
MAX_COLUMN_BYTE_SIZE: int = 50000000


class NodeType(int, Enum):
    """
    The types of Nodes we will see.

    The second nibble (4 bits) is a category marker, the first nibble is just an
    enumeration of the values in that category.

    This allows us to use bitmasks to add a category to the enumerations.
    """

    # fmt:off

    # 00000000
    UNKNOWN: int = 0

    # LOGICAL OPERATORS
    # 0001 nnnn
    AND: int = 17  # 0001 0001
    OR: int = 18  # 0001 0010 
    XOR: int = 19  # 0001 0011
    NOT: int = 20  # 0001 0100

    # INTERAL IDENTIFIERS
    # 0010 nnnn
    WILDCARD: int = 33  # 0010 0001
    COMPARISON_OPERATOR: int = 34  # 0010 0010
    BINARY_OPERATOR: int = 35  # 0010 0011
    UNARY_OPERATOR: int = 36  # 0010 0100
    FUNCTION: int = 37  # 0010 0101
    IDENTIFIER: int = 38  # 0010 0110
    SUBQUERY: int = 39  # 0010 0111
    NESTED: int = 40  # 0010 1000
    AGGREGATOR:int = 41  # 0010 1001
    LITERAL:int = 42  # 0010 1010
    EXPRESSION_LIST: int = 43  # 0010 1011 (CASE WHEN)
    EVALUATED: int = 44  # 0010 1100 - memoize results


ORSO_TO_NUMPY_MAP = {
    OrsoTypes.ARRAY: numpy.dtype("O"),
    OrsoTypes.BLOB: numpy.dtype("S"),
    OrsoTypes.BOOLEAN: numpy.dtype("?"),
    OrsoTypes.DATE: numpy.dtype("datetime64[D]"),  # [2.5e16 BC, 2.5e16 AD]
    OrsoTypes.DECIMAL: numpy.dtype("O"),
    OrsoTypes.DOUBLE: numpy.dtype("float64"),
    OrsoTypes.INTEGER: numpy.dtype("int64"),
    OrsoTypes.INTERVAL: numpy.dtype("m"),
    OrsoTypes.STRUCT: numpy.dtype("O"),
    OrsoTypes.TIMESTAMP: numpy.dtype("datetime64[us]"),  # [290301 BC, 294241 AD]
    OrsoTypes.TIME: numpy.dtype("O"),
    OrsoTypes.VARCHAR: numpy.unicode_(),
    OrsoTypes.NULL: numpy.dtype("O"),
}

LOGICAL_OPERATIONS: Dict[NodeType, Callable] = {
    NodeType.AND: pyarrow.compute.and_,
    NodeType.OR: pyarrow.compute.or_,
    NodeType.XOR: pyarrow.compute.xor,
}


class ExecutionContext:
    def __init__(self):
        self.results = {}

    def store(self, identity, result):
        self.results[identity] = result

    def retrieve(self, identity):
        return self.results.get(identity)

    def has_result(self, identity):
        return identity in self.results

    @property
    def identities(self):
        return list(self.results.keys())


def short_cut_and(root, table, context):
    # Convert to NumPy arrays
    true_indices = numpy.arange(table.num_rows)

    # Evaluate left expression
    left_result = numpy.array(evaluate(root.left, table, context))
    left_result = numpy.asarray(left_result, dtype=bool)

    # If all values in left_result are False, no need to evaluate the right expression
    if not left_result.any():
        return left_result

    # Filter out indices where left_result is FALSE
    subset_indices = true_indices[left_result]

    # Create a subset table for evaluating the right expression
    subset_table = table.take(subset_indices)

    # Evaluate right expression on the subset table
    right_result = numpy.array(evaluate(root.right, subset_table, None))

    # Combine results
    # Iterate over subset_indices and update left_result at those positions
    left_result[subset_indices] = right_result

    return left_result


def short_cut_or(root, table, context):
    # Assuming table.num_rows returns the number of rows in the table
    false_indices = numpy.arange(table.num_rows)

    # Evaluate left expression
    left_result = numpy.array(evaluate(root.left, table, context), dtype=numpy.bool_)

    # Filter out indices where left_result is TRUE
    subset_indices = false_indices[~left_result]

    if subset_indices.size == 0:
        return left_result

    # Create a subset table for evaluating the right expression
    subset_table = table.take(subset_indices)

    # Evaluate right expression on the subset table
    right_result = numpy.array(evaluate(root.right, subset_table, None), dtype=numpy.bool_)

    # Combine results
    # Update left_result with the right_result where left_result was False
    left_result[subset_indices] = left_result[subset_indices] | right_result

    return left_result


def prioritize_evaluation(expressions):
    non_dependent_expressions = []
    dependent_expressions = []

    for expression in expressions:
        if not get_all_nodes_of_type(expression, (NodeType.EVALUATED,)):
            non_dependent_expressions.append(expression)
        else:
            dependent_expressions.append(expression)

    # Now that we have split the expressions into non-dependent and dependent,
    # we can return them in the desired order of evaluation.
    return non_dependent_expressions + dependent_expressions


def _inner_evaluate(root: Node, table: Table, context: ExecutionContext):
    node_type = root.node_type  # type:ignore

    if node_type == NodeType.SUBQUERY:
        raise UnsupportedSyntaxError("IN (<subquery>) temporarily not supported.")

    identity = root.schema_column.identity if root.schema_column else random_string()

    # If already evaluated, return memoized result.
    if context.has_result(identity):
        return context.retrieve(identity)

    # if we have this column already, just return it
    if identity in table.column_names:
        return table[identity].to_numpy()

    # LITERAL TYPES
    if node_type == NodeType.LITERAL:
        # if it's a literal value, return it once for every value in the table
        literal_type = root.type
        if literal_type == OrsoTypes.ARRAY:
            # this isn't as fast as .full - but lists and strings are problematic
            return numpy.array([root.value] * table.num_rows)
        if literal_type == OrsoTypes.VARCHAR:
            return numpy.array([root.value] * table.num_rows, dtype=numpy.unicode_)
        if literal_type == OrsoTypes.INTERVAL:
            return pyarrow.array([root.value] * table.num_rows)
        return numpy.full(
            shape=table.num_rows,
            fill_value=root.value,
            dtype=ORSO_TO_NUMPY_MAP[literal_type],
        )  # type:ignore

    # BOOLEAN OPERATORS
    if node_type & LOGICAL_TYPE == LOGICAL_TYPE:  # type:ignore
        if node_type == NodeType.OR:
            return short_cut_or(root, table, context)
        if node_type == NodeType.AND:
            return short_cut_and(root, table, context)

        if node_type in LOGICAL_OPERATIONS:
            left = _inner_evaluate(root.left, table, context) if root.left else [None]
            right = _inner_evaluate(root.right, table, context) if root.right else [None]
            return LOGICAL_OPERATIONS[node_type](left, right)  # type:ignore

        if node_type == NodeType.NOT:
            centre = _inner_evaluate(root.centre, table, context) if root.centre else [None]
            centre = pyarrow.array(centre, type=pyarrow.bool_())
            return pyarrow.compute.invert(centre)

    # INTERAL IDENTIFIERS
    if node_type & INTERNAL_TYPE == INTERNAL_TYPE:  # type:ignore
        if node_type == NodeType.FUNCTION:
            parameters = [_inner_evaluate(param, table, context) for param in root.parameters]
            # zero parameter functions get the number of rows as the parameter
            if len(parameters) == 0:
                parameters = [table.num_rows]
            func = FUNCTIONS[root.value]
            result = func(*parameters)
            if isinstance(result, list):
                result = numpy.array(result)
            context.store(identity, result)
            return result
        if node_type == NodeType.AGGREGATOR:
            # detected as an aggregator, but here it's an identifier because it
            # will have already been evaluated
            node_type = NodeType.EVALUATED
            root.value = format_expression(root)
            root.node_type = NodeType.EVALUATED
        if node_type == NodeType.EVALUATED:
            if root.schema_column.identity not in table.column_names:
                raise ColumnReferencedBeforeEvaluationError(column=root.schema_column.name)
            return table[root.schema_column.identity].to_numpy()
        if node_type == NodeType.COMPARISON_OPERATOR:
            left = _inner_evaluate(root.left, table, context)
            right = _inner_evaluate(root.right, table, context)
            result = filter_operations(
                left,
                root.left.schema_column.type,
                root.value,
                right,
                root.right.schema_column.type,
            )
            context.store(identity, result)
            return result
        if node_type == NodeType.BINARY_OPERATOR:
            left = _inner_evaluate(root.left, table, context)
            right = _inner_evaluate(root.right, table, context)
            result = binary_operations(
                left,
                root.left.schema_column.type,
                root.value,
                right,
                root.right.schema_column.type,
            )
            context.store(identity, result)
            return result
        if node_type == NodeType.WILDCARD:
            numpy.full(table.num_rows, "*", dtype=numpy.unicode_)
        if node_type == NodeType.SUBQUERY:
            # we should have a query plan here
            sub = root.value.execute()
            return pyarrow.concat_tables(sub, promote_options="none")
        if node_type == NodeType.NESTED:
            return _inner_evaluate(root.centre, table, context)
        if node_type == NodeType.UNARY_OPERATOR:
            centre = _inner_evaluate(root.centre, table, context)
            result = UNARY_OPERATIONS[root.value](centre)
            context.store(identity, result)
            return result
        if node_type == NodeType.EXPRESSION_LIST:
            values = [_inner_evaluate(val, table, context) for val in root.parameters]
            return values
        from opteryx.exceptions import ColumnNotFoundError

        raise ColumnNotFoundError(
            message=f"Unable to locate column '{root.source_column}' this is likely due to differences in SELECT and GROUP BY clauses."
        )


def evaluate(expression: Node, table: Table, context: Optional[ExecutionContext] = None):
    if context is None:
        context = ExecutionContext()

    result = _inner_evaluate(root=expression, table=table, context=context)
    if not isinstance(result, (pyarrow.Array, numpy.ndarray)):
        result = numpy.array(result)
    return result


def get_all_nodes_of_type(root, select_nodes):
    """
    Walk a expression tree collecting all the nodes of a specified type.
    """
    if root is None:
        return []
    if not isinstance(root, (set, tuple, list)):
        root = [root]

    identifiers = []
    for node in root:
        if node.node_type in select_nodes:
            identifiers.append(node)
        if node.left:
            identifiers.extend(get_all_nodes_of_type(node.left, select_nodes))
        if node.centre:
            identifiers.extend(get_all_nodes_of_type(node.centre, select_nodes))
        if node.right:
            identifiers.extend(get_all_nodes_of_type(node.right, select_nodes))
        if node.parameters:
            for parameter in node.parameters:
                if isinstance(parameter, (Node, LogicalColumn)):
                    identifiers.extend(get_all_nodes_of_type(parameter, select_nodes))

    return identifiers


def evaluate_and_append(expressions, table: Table):
    """
    Evaluate an expression and add it to the table.

    This needs to be able to deal with and avoid cascading problems where field names
    are duplicated, this is most common when performing many joins on the same table.
    """
    prioritized_expressions = prioritize_evaluation(expressions)

    context = ExecutionContext()

    for statement in prioritized_expressions:
        if statement.schema_column.identity in table.column_names:
            continue

        if should_evaluate(statement):
            new_column = evaluate_statement(statement, table, context)

            if isinstance(new_column, pyarrow.ChunkedArray):
                new_column = new_column.combine_chunks()
            table = table.append_column(statement.schema_column.identity, new_column)
            for identity in context.identities:
                if identity not in table.column_names:
                    table = table.append_column(identity, [context.retrieve(identity)])

    return table


def should_evaluate(statement):
    """Determine if the given statement should be evaluated."""
    valid_node_types = {
        NodeType.FUNCTION,
        NodeType.BINARY_OPERATOR,
        NodeType.COMPARISON_OPERATOR,
        NodeType.UNARY_OPERATOR,
        NodeType.NESTED,
        NodeType.NOT,
        NodeType.AND,
        NodeType.OR,
        NodeType.XOR,
        NodeType.LITERAL,
    }
    return statement.node_type in valid_node_types


def evaluate_statement(statement, table, context):
    """Evaluate a statement and return the corresponding column."""
    new_column = evaluate(statement, table, context)
    if is_mask(new_column, statement, table):
        new_column = create_mask(new_column, table.num_rows)
    return format_column(new_column)


def is_mask(new_column, statement, table):
    """Determine if the given column represents a mask."""
    return len(new_column) < table.num_rows or statement.node_type == NodeType.UNARY_OPERATOR


def create_mask(column, num_rows):
    """Create a boolean mask based on the given column."""
    bool_list = numpy.full(num_rows, False)
    bool_list[column] = True
    return bool_list


def format_column(column):
    """Format the column based on its size."""
    if column.nbytes > MAX_COLUMN_BYTE_SIZE:
        return [[i] for i in column]
    return [column]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This optimization runs toward the end of the set, it removes operators which
were useful during planning and optimization.

- Some projections are redundant (relecting down to the columns which the
  providing operation has already limited down to).
- SubQuery nodes are useful for planning and optimization, but don't do
  anything during execution, we can remove them here.

Both of these operations are cheap to execute, the benefit for this
optimization isn't expected to be realized until we implement multiprocessing
and there is work associated with IPC which we are avoiding by removing
impotent steps.
"""

from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext


class RedundantOperationsStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        # If we're a project and the providing step has the same columns, we're
        # not doing anything so can be removed.
        if node.node_type == LogicalPlanStepType.Project:
            providers = context.pre_optimized_tree.ingoing_edges(context.node_id)
            if len(providers) == 1:
                provider_node = context.pre_optimized_tree[providers[0][0]]
                provider_columns = {c.schema_column.identity for c in provider_node.columns}
                my_columns = {c.schema_column.identity for c in node.columns}
                if provider_columns == my_columns:
                    context.optimized_plan.remove_node(context.node_id, heal=True)

        # Subqueries are useful for planning but not needed for execution
        if node.node_type == LogicalPlanStepType.Subquery:
            context.optimized_plan.remove_node(context.node_id, heal=True)

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Global Buffer Pool.

This is uses an LRU-K2 policy to determine what to keep and what to evict and is
backed by a MemoryPool.

The buffer pool is has no slot limit, it is a given volume of memory, the pool
will try to evict when full. This is different to a classic Buffer Pool which
is slot-based.

The Buffer Pool is a global resource and used across all Connections and Cursors.
"""

from typing import Optional

from opteryx.config import MAX_LOCAL_BUFFER_CAPACITY
from opteryx.shared import MemoryPool
from opteryx.utils.lru_2 import LRU2


class _BufferPool:
    """
    Buffer Pool is a class implementing a Least Recently Used (LRU) policy for
    eviction.
    """

    slots = "_lru", "_memory_pool", "size"

    def __init__(self):
        self._lru = LRU2()
        self._memory_pool = MemoryPool(name="BufferPool", size=MAX_LOCAL_BUFFER_CAPACITY)
        self.size = self._memory_pool.size

    def get(self, key: bytes, zero_copy: bool = True) -> Optional[bytes]:
        """
        Retrieve an item from the pool, return None if the item isn't found.
        """
        mp_key = self._lru.get(key)
        if mp_key is not None:
            return self._memory_pool.read(mp_key, zero_copy=zero_copy)
        return None

    def set(self, key: bytes, value) -> Optional[str]:
        """
        Attempt to save a value to the buffer pool. Check first if there is space to commit the value.
        If not, evict the least recently used item and try again.

        Args:
            key: The key associated with the value to commit.
            value: The value to commit to the buffer pool.

        Returns:
            The key of the evicted item if eviction occurred, otherwise None.
        """
        # First check if we can commit the value to the memory pool
        if not self._memory_pool.available_space() >= len(value):
            evicted_key, evicted_value = self._lru.evict(details=True)
            if evicted_key:
                self._memory_pool.release(evicted_value)
            else:
                return None  # Return None if no item could be evicted

        # Try to commit the value to the memory pool
        memory_pool_key = self._memory_pool.commit(value)
        if memory_pool_key is None:
            return None  # Return None if commit still fails after eviction

        # Update LRU cache with the new key and memory pool key if commit succeeds
        self._lru.set(key, memory_pool_key)

        # Return the evicted key if an eviction occurred, otherwise return None
        return evicted_key if "evicted_key" in locals() else None

    @property
    def stats(self) -> tuple:
        """
        Return the hit, miss and eviction statistics for the buffer pool.
        """
        return self._lru.stats

    def reset(self, reset_stats: bool = False):
        """
        Reset the buffer pool.
        If reset_stats is True, also reset the statistics.
        """
        self._lru.reset(reset_stats=reset_stats)


class BufferPool(_BufferPool):
    """
    Singleton wrapper for the _BufferPool class. Only allows one instance of _BufferPool to exist.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = cls._create_instance()
        return cls._instance

    @classmethod
    def _create_instance(cls):
        return _BufferPool()

    @classmethod
    def reset(cls):
        """
        Reset the BufferPool singleton instance. This is useful when the configuration changes.
        """
        cls._instance = None
        cls._instance = cls._create_instance()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy
import pyarrow
from orso.types import OrsoTypes
from pyarrow import compute

# fmt:off
OPERATOR_FUNCTION_MAP: Dict[str, Any] = {
    "Divide": numpy.divide,
    "Minus": numpy.subtract,
    "Modulo": numpy.mod,
    "Multiply": numpy.multiply,
    "Plus": numpy.add,
    "StringConcat": compute.binary_join_element_wise,
    "MyIntegerDivide": lambda left, right: numpy.trunc(numpy.divide(left, right)).astype(numpy.int64),
    "BitwiseOr": numpy.bitwise_or,
    "BitwiseAnd": numpy.bitwise_and,
    "BitwiseXor": numpy.bitwise_xor,
    "ShiftLeft": numpy.left_shift,
    "ShiftRight": numpy.right_shift,
}

BINARY_OPERATORS = set(OPERATOR_FUNCTION_MAP.keys())

# fmt:on


def _ip_containment(left: List[Optional[str]], right: List[str]) -> List[Optional[bool]]:
    """
    Check if each IP address in 'left' is contained within the network specified in 'right'.

    Parameters:
        left: List[Optional[str]]
            List of IP addresses as strings.
        right: List[str]
            List containing the network as a string.

    Returns:
        List[Optional[bool]]:
            A list of boolean values indicating if each corresponding IP in 'left' is in 'right'.
    """

    from opteryx.compiled.functions import ip_in_cidr

    try:
        return ip_in_cidr(left, str(right[0]))
    except (IndexError, AttributeError, ValueError) as err:
        from opteryx.exceptions import IncorrectTypeError

        raise IncorrectTypeError(
            "The `|` operator can be used as bitwise OR or IP address containment only."
        ) from err


def binary_operations(
    left, left_type: OrsoTypes, operator: str, right, right_type: OrsoTypes
) -> Union[numpy.ndarray, pyarrow.Array]:
    """
    Execute inline operators (e.g. the add in 3 + 4).

    Parameters:
        left: Union[numpy.ndarray, pyarrow.Array]
            The left operand
        operator: str
            The operator to be applied
        right: Union[numpy.ndarray, pyarrow.Array]
            The right operand
    Returns:
        Union[numpy.ndarray, pyarrow.Array]
            The result of the binary operation
    """
    operation = OPERATOR_FUNCTION_MAP.get(operator)

    if operation is None:
        raise NotImplementedError(f"Operator `{operator}` is not implemented!")

    if OrsoTypes.INTERVAL in (left_type, right_type):
        from opteryx.custom_types.intervals import INTERVAL_KERNELS

        function = INTERVAL_KERNELS.get((left_type, right_type, operator))
        if function is None:
            from opteryx.exceptions import UnsupportedTypeError

            raise UnsupportedTypeError(
                f"Cannot perform {operator.upper()} on {left_type} and {right_type}."
            )

        return function(left, left_type, right, right_type, operator)

    if (
        operator == "Minus"
        and left_type in (OrsoTypes.DATE, OrsoTypes.TIMESTAMP)
        and right_type in (OrsoTypes.DATE, OrsoTypes.TIMESTAMP)
    ):
        # substracting dates results in an INTERVAL (months, seconds)
        arr = operation(left, right)
        if arr.dtype.name == "timedelta64[D]":
            return pyarrow.array(
                [
                    None if v == -9223372036854775808 else (0, v * 86400)
                    for v in arr.astype(numpy.int64)
                ]
            )
        arr = arr.astype("timedelta64[s]").astype(numpy.int64)
        return pyarrow.array([(0, v) for v in arr.astype(numpy.int64)])

    elif operator == "BitwiseOr":
        if OrsoTypes.VARCHAR in (left_type, right_type):
            return _ip_containment(left, right)

    elif operator == "StringConcat":
        empty = numpy.full(len(left), "")
        joined = compute.binary_join_element_wise(left, right, empty)
        return joined

    return operation(left, right)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Decode files from a raw binary format to a PyArrow Table.
"""

import io
from enum import Enum
from typing import BinaryIO
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

import numpy
import pyarrow
from orso.tools import random_string

from opteryx.exceptions import UnsupportedFileTypeError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.utils.arrow import post_read_projector
from opteryx.utils.memory_view_stream import MemoryViewStream


class ExtentionType(str, Enum):
    """labels for the file extentions"""

    DATA = "DATA"
    CONTROL = "CONTROL"


def convert_arrow_schema_to_orso_schema(arrow_schema):
    from orso.schema import FlatColumn
    from orso.schema import RelationSchema

    return RelationSchema(
        name="arrow",
        columns=[FlatColumn.from_arrow(field) for field in arrow_schema],
    )


def get_decoder(dataset: str) -> Callable:
    """helper routine to get the decoder for a given file"""
    ext = dataset.split(".")[-1].lower()
    if ext not in KNOWN_EXTENSIONS:  # pragma: no cover
        raise UnsupportedFileTypeError(f"Unsupported file type - {ext}")
    file_decoder, file_type = KNOWN_EXTENSIONS[ext]
    if file_type != ExtentionType.DATA:  # pragma: no cover
        raise UnsupportedFileTypeError(f"File is not a data file - {ext}")
    return file_decoder


def do_nothing(buffer, projection=None, just_schema: bool = False):  # pragma: no cover
    """for when you need to look like you're doing something"""
    return False


def filter_records(filter, table):
    """
    When we can't push predicates to the actual read, use this to filter records
    just after the read.
    """
    # notes:
    #   at this point we've not renamed any columns, this may affect some filters
    from opteryx.managers.expression import NodeType
    from opteryx.managers.expression import evaluate
    from opteryx.models import Node

    if isinstance(filter, list) and filter:
        filter_copy = [p for p in filter]
        root = filter_copy.pop()

        if root.left.node_type == NodeType.IDENTIFIER:
            root.left.schema_column.identity = root.left.source_column
        if root.right.node_type == NodeType.IDENTIFIER:
            root.right.schema_column.identity = root.right.source_column

        while filter_copy:
            right = filter_copy.pop()
            if right.left.node_type == NodeType.IDENTIFIER:
                right.left.schema_column.identity = right.left.source_column
            if right.right.node_type == NodeType.IDENTIFIER:
                right.right.schema_column.identity = right.right.source_column
            root = Node(
                NodeType.AND,
                left=root,
                right=right,
                schema_column=Node("schema_column", identity=random_string()),
            )
    else:
        root = filter

    mask = evaluate(root, table)
    return table.filter(mask)


def zstd_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    """
    Read zstandard compressed JSONL files
    """
    import zstandard

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)

    with zstandard.open(stream, "rb") as file:
        return jsonl_decoder(
            file, projection=projection, selection=selection, just_schema=just_schema
        )


def lzma_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    """
    Read lzma compressed JSONL files
    """
    import lzma

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)

    with lzma.open(stream, "rb") as file:
        return jsonl_decoder(
            file, projection=projection, selection=selection, just_schema=just_schema
        )


def parquet_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    """
    Read parquet formatted files.
    """
    from pyarrow import parquet

    from opteryx.connectors.capabilities import PredicatePushable

    columns_in_filters = {c.value for c in get_all_nodes_of_type(selection, (NodeType.IDENTIFIER,))}

    # Convert the selection to DNF format if applicable
    dnf_filter, selection = PredicatePushable.to_dnf(selection) if selection else (None, None)
    selected_columns = None

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)

    # Open the parquet file only once
    parquet_file = parquet.ParquetFile(stream)
    if projection or just_schema:
        # Return just the schema if that's all that's needed
        if just_schema:
            return convert_arrow_schema_to_orso_schema(parquet_file.schema_arrow)

        # Projection processing
        schema_columns_set = set(parquet_file.schema_arrow.names)
        projection_names = {
            name for proj_col in projection for name in proj_col.schema_column.all_names
        }
        selected_columns = list(schema_columns_set.intersection(projection_names))

        # If no columns are selected, set to None
        if not selected_columns:
            selected_columns = None
        if not columns_in_filters.issubset(schema_columns_set):
            if selected_columns is None:
                selected_columns = list(schema_columns_set)
            fields = [pyarrow.field(name, pyarrow.string()) for name in selected_columns]
            schema = pyarrow.schema(fields)

            # Create an empty table with the schema
            empty_table = pyarrow.Table.from_arrays(
                [pyarrow.array([], type=schema.field(i).type) for i in range(len(fields))],
                schema=schema,
            )

            return (
                parquet_file.metadata.num_rows,
                parquet_file.metadata.num_columns,
                empty_table,
            )

    # Special handling for projection of [] (usually means COUNT(*))
    if projection == []:
        selected_columns = []

    # Read the parquet table with the optimized column list and selection filters
    table = parquet.read_table(
        stream,
        columns=selected_columns,
        pre_buffer=False,
        filters=dnf_filter,
        use_threads=False,
    )
    if selection:
        table = filter_records(selection, table)
    if projection == []:
        return (
            parquet_file.metadata.num_rows,
            parquet_file.metadata.num_columns,
            pyarrow.Table.from_pydict({"_": numpy.full(table.num_rows, True, dtype=numpy.bool_)}),
        )
    return (parquet_file.metadata.num_rows, parquet_file.metadata.num_columns, table)


def orc_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    """
    Read orc formatted files
    """
    import pyarrow.orc as orc

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)
    orc_file = orc.ORCFile(stream)

    if just_schema:
        orc_schema = orc_file.schema
        return convert_arrow_schema_to_orso_schema(orc_schema)

    table = orc_file.read()
    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)
    return *full_shape, table


def jsonl_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    import pyarrow.json

    stream: BinaryIO = None
    if isinstance(buffer, memoryview):
        stream = MemoryViewStream(buffer)
    elif isinstance(buffer, bytes):
        stream = io.BytesIO(buffer)
    else:
        stream = buffer

    table = pyarrow.json.read_json(stream)
    schema = table.schema
    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, table


def csv_decoder(
    buffer,
    projection: List = None,
    selection=None,
    delimiter: str = ",",
    just_schema: bool = False,
):
    import pyarrow.csv
    from pyarrow.csv import ParseOptions

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)
    parse_options = ParseOptions(delimiter=delimiter, newlines_in_values=True)
    table = pyarrow.csv.read_csv(stream, parse_options=parse_options)
    schema = table.schema
    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, table


def tsv_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    return csv_decoder(
        buffer=buffer,
        projection=projection,
        selection=selection,
        delimiter="\t",
        just_schema=just_schema,
    )


def arrow_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    import pyarrow.feather as pf

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)
    table = pf.read_table(stream)
    schema = table.schema
    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, table


def avro_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    """
    AVRO isn't well supported, it is converted between table types which is
    really slow
    """
    try:
        from avro.datafile import DataFileReader
        from avro.io import DatumReader
    except ImportError:  # pragma: no cover
        raise Exception("`avro` is missing, please install or include in your `requirements.txt`.")

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)
    reader = DataFileReader(stream, DatumReader())

    table = pyarrow.Table.from_pylist(list(reader))
    schema = table.schema
    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, table


def ipc_decoder(buffer, projection: List = None, selection=None, just_schema: bool = False):
    from itertools import chain

    from pyarrow import ipc

    stream: BinaryIO = None
    stream = MemoryViewStream(buffer) if isinstance(buffer, memoryview) else io.BytesIO(buffer)
    reader = ipc.open_stream(stream)

    batch_one = next(reader, None)
    if batch_one is None:
        return None

    schema = batch_one.schema

    if just_schema:
        return convert_arrow_schema_to_orso_schema(schema)

    table = pyarrow.Table.from_batches([batch for batch in chain([batch_one], reader)])
    full_shape = table.shape
    if selection:
        table = filter_records(selection, table)
    if projection:
        table = post_read_projector(table, projection)

    return *full_shape, table


# for types we know about, set up how we handle them
KNOWN_EXTENSIONS: Dict[str, Tuple[Callable, str]] = {
    "avro": (avro_decoder, ExtentionType.DATA),
    "complete": (do_nothing, ExtentionType.CONTROL),
    "manifest": (do_nothing, ExtentionType.CONTROL),
    "ignore": (do_nothing, ExtentionType.CONTROL),
    "arrow": (arrow_decoder, ExtentionType.DATA),  # feather
    "csv": (csv_decoder, ExtentionType.DATA),
    "ipc": (ipc_decoder, ExtentionType.DATA),
    "jsonl": (jsonl_decoder, ExtentionType.DATA),
    "orc": (orc_decoder, ExtentionType.DATA),
    "parquet": (parquet_decoder, ExtentionType.DATA),
    "tsv": (tsv_decoder, ExtentionType.DATA),
    "zstd": (zstd_decoder, ExtentionType.DATA),  # jsonl/zstd
    "lzma": (lzma_decoder, ExtentionType.DATA),  # jsonl/lzma
}

VALID_EXTENSIONS = set(f".{ext}" for ext in KNOWN_EXTENSIONS)
TUPLE_OF_VALID_EXTENSIONS = tuple(VALID_EXTENSIONS)
DATA_EXTENSIONS = set(
    f".{ext}" for ext, conf in KNOWN_EXTENSIONS.items() if conf[1] == ExtentionType.DATA
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Stop Words - words not included in the VECTOR INDEX or in MATCH AGAINST searches
"""

from typing import Set

from orso.schema import FlatColumn
from orso.schema import RelationSchema
from orso.types import OrsoTypes

__all__ = ("read", "schema")

STOP_WORDS: Set[bytes] = {
    b"about",
    b"above",
    b"across",
    b"after",
    b"afterwards",
    b"again",
    b"against",
    b"all",
    b"almost",
    b"alone",
    b"along",
    b"already",
    b"also",
    b"although",
    b"always",
    b"am",
    b"among",
    b"amongst",
    b"amount",
    b"an",
    b"and",
    b"another",
    b"any",
    b"anyhow",
    b"anyone",
    b"anything",
    b"anyway",
    b"anywhere",
    b"are",
    b"around",
    b"as",
    b"at",
    b"back",
    b"be",
    b"became",
    b"because",
    b"become",
    b"becomes",
    b"becoming",
    b"been",
    b"before",
    b"beforehand",
    b"behind",
    b"being",
    b"below",
    b"beside",
    b"besides",
    b"between",
    b"beyond",
    b"both",
    b"bottom",
    b"but",
    b"by",
    b"ca",
    b"call",
    b"can",
    b"cannot",
    b"could",
    b"did",
    b"do",
    b"does",
    b"doing",
    b"done",
    b"down",
    b"due",
    b"during",
    b"each",
    b"eight",
    b"either",
    b"eleven",
    b"else",
    b"elsewhere",
    b"empty",
    b"enough",
    b"even",
    b"ever",
    b"every",
    b"everyone",
    b"everything",
    b"everywhere",
    b"except",
    b"few",
    b"fifteen",
    b"fifty",
    b"first",
    b"five",
    b"for",
    b"former",
    b"formerly",
    b"forty",
    b"four",
    b"from",
    b"front",
    b"full",
    b"further",
    b"get",
    b"give",
    b"go",
    b"had",
    b"has",
    b"have",
    b"he",
    b"hence",
    b"her",
    b"here",
    b"hereafter",
    b"hereby",
    b"herein",
    b"hereupon",
    b"hers",
    b"herself",
    b"him",
    b"himself",
    b"his",
    b"how",
    b"however",
    b"hundred",
    b"if",
    b"in",
    b"indeed",
    b"into",
    b"is",
    b"it",
    b"its",
    b"itself",
    b"just",
    b"keep",
    b"last",
    b"latter",
    b"latterly",
    b"least",
    b"less",
    b"ll",
    b"made",
    b"make",
    b"many",
    b"may",
    b"me",
    b"meanwhile",
    b"might",
    b"mine",
    b"more",
    b"moreover",
    b"most",
    b"mostly",
    b"move",
    b"much",
    b"must",
    b"my",
    b"myself",
    b"name",
    b"namely",
    b"neither",
    b"never",
    b"nevertheless",
    b"next",
    b"nine",
    b"no",
    b"nobody",
    b"none",
    b"noone",
    b"nor",
    b"not",
    b"nothing",
    b"now",
    b"nowhere",
    b"of",
    b"off",
    b"often",
    b"on",
    b"once",
    b"one",
    b"only",
    b"onto",
    b"or",
    b"other",
    b"others",
    b"otherwise",
    b"our",
    b"ours",
    b"ourselves",
    b"out",
    b"over",
    b"own",
    b"part",
    b"per",
    b"perhaps",
    b"please",
    b"put",
    b"quite",
    b"rather",
    b"re",
    b"really",
    b"regarding",
    b"same",
    b"say",
    b"see",
    b"seem",
    b"seemed",
    b"seeming",
    b"seems",
    b"serious",
    b"several",
    b"she",
    b"should",
    b"show",
    b"side",
    b"since",
    b"six",
    b"sixty",
    b"so",
    b"some",
    b"somehow",
    b"someone",
    b"something",
    b"sometime",
    b"sometimes",
    b"somewhere",
    b"still",
    b"such",
    b"take",
    b"ten",
    b"than",
    b"that",
    b"the",
    b"their",
    b"them",
    b"themselves",
    b"then",
    b"thence",
    b"there",
    b"thereafter",
    b"thereby",
    b"therefore",
    b"therein",
    b"thereupon",
    b"these",
    b"they",
    b"third",
    b"this",
    b"those",
    b"though",
    b"three",
    b"through",
    b"throughout",
    b"thru",
    b"thus",
    b"to",
    b"together",
    b"too",
    b"top",
    b"toward",
    b"towards",
    b"twelve",
    b"twenty",
    b"two",
    b"under",
    b"unless",
    b"until",
    b"up",
    b"upon",
    b"us",
    b"used",
    b"using",
    b"various",
    b"very",
    b"via",
    b"ve",
    b"was",
    b"we",
    b"well",
    b"were",
    b"what",
    b"whatever",
    b"when",
    b"whence",
    b"whenever",
    b"where",
    b"whereafter",
    b"whereas",
    b"whereby",
    b"wherein",
    b"whereupon",
    b"wherever",
    b"whether",
    b"which",
    b"while",
    b"whither",
    b"who",
    b"whoever",
    b"whole",
    b"whom",
    b"whose",
    b"why",
    b"will",
    b"with",
    b"within",
    b"without",
    b"would",
    b"yet",
    b"you",
    b"your",
    b"yours",
    b"yourself",
    b"yourselves",
}


def read(end_date=None, *args):
    import pyarrow

    # Define the data
    data = [
        pyarrow.array(
            STOP_WORDS,
            type=pyarrow.string(),
        ),
    ]
    column_names = ["value"]

    return pyarrow.Table.from_arrays(data, column_names)


def schema():
    # fmt:off
    return RelationSchema(
            name="$stop_words",
            columns=[
                FlatColumn(name="value", type=OrsoTypes.VARCHAR),
            ],
        )
    # fmt:on
